# -*- encoding: utf-8 -*-
"""
KERIA
keria.app.exchanging module

"""
import json

import falcon
from keri.core import coring, eventing, serdering
from keri.peer import exchanging

from keria.core import httping


def loadEnds(app):
    exnColEnd = ExchangeCollectionEnd()
    app.add_route("/identifiers/{name}/exchanges", exnColEnd)

    exnColEnd = ExchangeQueryCollectionEnd()
    app.add_route("/exchanges/query", exnColEnd)

    exnResEnd = ExchangeResourceEnd()
    app.add_route("/exchanges/{said}", exnResEnd)


class ExchangeCollectionEnd:

    @staticmethod
    def on_post(req, rep, name):
        """  POST endpoint for exchange message collection

        Args:
            req (Request): falcon HTTP request object
            rep (Response): falcon HTTP response object
            name (str): human readable alias for AID context

        """
        agent = req.context.agent

        body = req.get_media()

        # Get the hab
        hab = agent.hby.habByName(name)
        if hab is None:
            raise falcon.HTTPNotFound(description=f"alias={name} is not a valid reference to an identifier")

        # Get the exn, sigs, additional attachments and recipients  from the request
        ked = httping.getRequiredParam(body, "exn")
        sigs = httping.getRequiredParam(body, "sigs")
        atc = httping.getRequiredParam(body, "atc")
        rec = httping.getRequiredParam(body, "rec")
        topic = httping.getRequiredParam(body, "tpc")

        for recp in rec:  # Have to verify we already know all the recipients.
            if recp not in agent.hby.kevers:
                raise falcon.HTTPBadRequest(description=f"attempt to send to unknown AID={recp}")

        # use that data to create th Serder and Sigers for the exn
        serder = serdering.SerderKERI(sad=ked)
        sigers = [coring.Siger(qb64=sig) for sig in sigs]

        # Now create the stream to send, need the signer seal
        kever = hab.kever
        seal = eventing.SealEvent(i=hab.pre, s="{:x}".format(kever.lastEst.s), d=kever.lastEst.d)

        ims = eventing.messagize(serder=serder, sigers=sigers, seal=seal)

        # Have to add the atc to the end... this will be Pathed signatures for embeds
        ims.extend(atc.encode("utf-8"))  # add the pathed attachments

        # make a copy and parse
        agent.hby.psr.parseOne(ims=bytearray(ims))

        msg = dict(said=serder.said, pre=hab.pre, rec=rec, topic=topic)

        agent.exchanges.append(msg)

        rep.status = falcon.HTTP_202
        rep.data = json.dumps(serder.ked).encode("utf-8")


class ExchangeQueryCollectionEnd:

    @staticmethod
    def on_post(req, rep):
        """  POST endpoint for exchange message collection

        Args:
            req (Request): falcon HTTP request object
            rep (Response): falcon HTTP response object

        """
        agent = req.context.agent

        try:
            body = req.get_media()
            if "filter" in body:
                filtr = body["filter"]
            else:
                filtr = {}

            if "sort" in body:
                sort = body["sort"]
            else:
                sort = None

            if "skip" in body:
                skip = body["skip"]
            else:
                skip = 0

            if "limit" in body:
                limit = body["limit"]
            else:
                limit = 25
        except falcon.HTTPError:
            filtr = {}
            sort = {}
            skip = 0
            limit = 25

        cur = agent.exnseeker.find(filtr=filtr, sort=sort, skip=skip, limit=limit)
        saids = [coring.Saider(qb64=said) for said in cur]

        exns = []
        for said in saids:
            serder, pathed = exchanging.cloneMessage(agent.hby, said.qb64)
            exns.append(dict(exn=serder.ked, pathed={k: v.decode("utf-8") for k, v in pathed.items()}))

        rep.status = falcon.HTTP_200
        rep.content_type = "application/json"
        rep.data = json.dumps(exns).encode("utf-8")


class ExchangeResourceEnd:
    """ Exchange message resource endpoint class """

    @staticmethod
    def on_get(req, rep, said):
        """GET endpoint for exchange message collection

        Args:
            req (Request): falcon HTTP request object
            rep (Response): falcon HTTP response object
            said (str): qb64 SAID of exchange message to retrieve

        """
        agent = req.context.agent
        serder, pathed = exchanging.cloneMessage(agent.hby, said)

        if serder is None:
            raise falcon.HTTPNotFound(description=f"SAID {said} does not match a verified EXN message")

        exn = dict(exn=serder.ked, pathed={k: v.decode("utf-8") for k, v in pathed.items()})
        rep.status = falcon.HTTP_200
        rep.content_type = "application/json"
        rep.data = json.dumps(exn).encode("utf-8")



from hio.base import doing
from keri import kering
from keri.app import forwarding, agenting, habbing
from keri.core import coring, serdering
from keri.db import dbing


class Sealer(doing.DoDoer):
    """
    Sends messages to Delegator of an identifier and wait for the anchoring event to
    be processed to ensure the inception or rotation event has been approved by the delegator.

    Removes all Doers and exits as Done once the event has been anchored.

    """

    def __init__(self, hby, proxy=None, **kwa):
        """
        For the current event, gather the current set of witnesses, send the event,
        gather all receipts and send them to all other witnesses

        Parameters:
            hab (Hab): Habitat of the identifier to populate witnesses
            msg (bytes): is the message to send to all witnesses.
                 Defaults to sending the latest KEL event if msg is None
            scheme (str): Scheme to favor if available

        """
        self.hby = hby
        self.postman = forwarding.Poster(hby=hby)
        self.witq = agenting.WitnessInquisitor(hby=hby)
        self.witDoer = agenting.Receiptor(hby=self.hby)
        self.proxy = proxy

        super(Sealer, self).__init__(doers=[self.witq, self.witDoer, self.postman, doing.doify(self.escrowDo)],
                                     **kwa)

    def delegation(self, pre, sn=None, proxy=None):
        if pre not in self.hby.habs:
            raise kering.ValidationError(f"{pre} is not a valid local AID for delegation")

        # load the hab of the delegated identifier to anchor
        hab = self.hby.habs[pre]
        delpre = hab.kever.delegator  # get the delegator identifier
        if delpre not in hab.kevers:
            raise kering.ValidationError(f"delegator {delpre} not found, unable to process delegation")

        sn = sn if sn is not None else hab.kever.sner.num

        # load the event and signatures
        evt = hab.makeOwnEvent(sn=sn)

        if isinstance(hab, habbing.GroupHab):
            phab = hab.mhab
        elif hab.kever.sn > 0:
            phab = hab
        elif proxy is not None:
            phab = proxy
        elif self.proxy is not None:
            phab = self.proxy
        else:
            raise kering.ValidationError("no proxy to send messages for delegation")

        srdr = serdering.SerderKERI(raw=evt)
        del evt[:srdr.size]
        self.postman.send(hab=phab, dest=delpre, topic="delegate", serder=srdr, attachment=evt)

        self.hby.db.dune.pin(keys=(srdr.pre, srdr.said), val=srdr)

    def complete(self, prefixer, seqner, saider=None):
        """ Check for completed delegation protocol for the specific event

        Parameters:
            prefixer (Prefixer): qb64 identifier prefix of event to check
            seqner (Seqner): sequence number of event to check
            saider (Saider): optional digest of event to verify

        Returns:

        """
        csaider = self.hby.db.cdel.get(keys=(prefixer.qb64, seqner.qb64))
        if not csaider:
            return False
        else:
            if saider and (csaider.qb64 != saider.qb64):
                raise kering.ValidationError(f"invalid delegation protocol escrowed event {csaider.qb64}-{saider.qb64}")

        return True

    def escrowDo(self, tymth, tock=1.0):
        """ Process escrows of group multisig identifiers waiting to be compeleted.

        Steps involve:
           1. Sending local event with sig to other participants
           2. Waiting for signature threshold to be met.
           3. If elected and delegated identifier, send complete event to delegator
           4. If delegated, wait for delegator's anchored seal
           5. If elected, send event to witnesses and collect receipts.
           6. Otherwise, wait for fully receipted event

        Parameters:
            tymth (function): injected function wrapper closure returned by .tymen() of
                Tymist instance. Calling tymth() returns associated Tymist .tyme.
            tock (float): injected initial tock value.  Default to 1.0 to slow down processing

        """
        # enter context
        self.wind(tymth)
        self.tock = tock
        _ = (yield self.tock)

        while True:
            self.processEscrows()
            yield 0.5

    def processEscrows(self):
        self.processUnanchoredEscrow()
        self.processPartialWitnessEscrow()

    def processUnanchoredEscrow(self):
        """
        Process escrow of partially signed multisig group KEL events.  Message
        processing will send this local controllers signature to all other participants
        then this escrow waits for signatures from all other participants

        """
        for (pre, said), serder in self.hby.db.dune.getItemIter():  # group partial witness escrow
            kever = self.hby.kevers[pre]
            dkever = self.hby.kevers[kever.delegator]

            seal = dict(i=serder.pre, s=serder.snh, d=serder.said)
            if dserder := self.hby.db.findAnchoringSealEvent(dkever.prefixer.qb64, seal=seal):
                seqner = coring.Seqner(sn=dserder.sn)
                couple = seqner.qb64b + dserder.saidb
                dgkey = dbing.dgKey(kever.prefixer.qb64b, kever.serder.saidb)
                self.hby.db.setAes(dgkey, couple)  # authorizer event seal (delegator/issuer)
                self.witDoer.msgs.append(dict(pre=pre, sn=serder.sn))

                # Move to escrow waiting for witness receipts
                print(f"Waiting for fully signed witness receipts for {serder.sn}")
                self.hby.db.dpwe.pin(keys=(pre, said), val=serder)
                self.hby.db.dune.rem(keys=(pre, said))

    def processPartialWitnessEscrow(self):
        """
        Process escrow of delegated events that do not have a full compliment of receipts
        from witnesses yet.  When receipting is complete, remove from escrow and cue up a message
        that the event is complete.

        """
        for (pre, said), serder in self.hby.db.dpwe.getItemIter():  # group partial witness escrow
            kever = self.hby.kevers[pre]
            dgkey = dbing.dgKey(pre, serder.said)
            seqner = coring.Seqner(sn=serder.sn)

            # Load all the witness receipts we have so far
            wigs = self.hby.db.getWigs(dgkey)
            if len(wigs) == len(kever.wits):  # We have all of them, this event is finished
                if len(kever.wits) > 0:
                    witnessed = False
                    for cue in self.witDoer.cues:
                        if cue["pre"] == serder.pre and cue["sn"] == seqner.sn:
                            witnessed = True
                    if not witnessed:
                        continue
                print(f"Witness receipts complete, {pre} confirmed.")
                self.hby.db.dpwe.rem(keys=(pre, said))
                self.hby.db.cdel.put(keys=(pre, seqner.qb64), val=coring.Saider(qb64=serder.said))



# -*- encoding: utf-8 -*-
"""
KERI
keria.core.keeping module

"""
from dataclasses import dataclass, asdict, field

from keri.app.keeping import PreSit, Algos, PubLot, PubSet
from keri.core import coring
from keri.core.coring import Tiers, MtrDex
from keri.db import dbing, subing, koming
from keri.help import helping

from keria import core


@dataclass()
class Prefix:
    pidx: int = 0  # prefix index for this keypair sequence
    algo: str = Algos.salty  # salty default uses indices and salt to create new key pairs


@dataclass()
class SaltyPrm:
    """
    Salty prefix's parameters for creating new key pairs
    """
    sxlt: str = ''  # qualified b64 encoded AID salt
    pidx: int = 0  # prefix index for this keypair sequence
    kidx: int = 0  # key index for this keypair sequence
    stem: str = ''  # default unique path stem for salty algo
    tier: str = ''  # security tier for stretch index salty algo
    dcode: str = ''  # next digest hasing code
    icodes: list = field(default_factory=list)  # current signing key seed codes
    ncodes: list = field(default_factory=list)  # next key seed codes
    transferable: bool = False

    def __iter__(self):
        return iter(asdict(self))


class RemoteKeeper(dbing.LMDBer):
    """
    RemoteKeeper stores data for Salty or Randy Encrypted edge key generation.

    """

    TailDirPath = "keri/rks"
    AltTailDirPath = ".keri/rks"
    TempPrefix = "keri_rks_"
    MaxNamedDBs = 10

    def __init__(self, headDirPath=None, perm=None, reopen=False, **kwa):
        """
        Setup named sub databases.

        Inherited Parameters:
            name is str directory path name differentiator for main database
                When system employs more than one keri database, name allows
                differentiating each instance by name
                default name='main'
            temp is boolean, assign to .temp
                True then open in temporary directory, clear on close
                Othewise then open persistent directory, do not clear on close
                default temp=False
            headDirPath is optional str head directory pathname for main database
                If not provided use default .HeadDirpath
                default headDirPath=None so uses self.HeadDirPath
            perm is numeric optional os dir permissions mode
                default perm=None so do not set mode
            reopen is boolean, IF True then database will be reopened by this init
                default reopen=True

        Notes:

        dupsort=True for sub DB means allow unique (key,pair) duplicates at a key.
        Duplicate means that is more than one value at a key but not a redundant
        copies a (key,value) pair per key. In other words the pair (key,value)
        must be unique both key and value in combination.
        Attempting to put the same (key,value) pair a second time does
        not add another copy.

        Duplicates are inserted in lexocographic order by value, insertion order.

        """
        self.pubs = None
        self.sits = None
        self.sprms = None
        self.pres = None
        self.mhabs = None
        self.nxts = None
        self.prxs = None
        self.gbls = None
        if perm is None:
            perm = self.Perm  # defaults to restricted permissions for non temp

        super(RemoteKeeper, self).__init__(headDirPath=headDirPath, perm=perm,
                                           reopen=reopen, **kwa)

    def reopen(self, **kwa):
        """
        Open sub databases
        """
        self.opened = super(RemoteKeeper, self).reopen(**kwa)

        # Create by opening first time named sub DBs within main DB instance
        # Names end with "." as sub DB name must include a non Base64 character
        # to avoid namespace collisions with Base64 identifier prefixes.

        self.gbls = subing.Suber(db=self, subkey='gbls.')
        self.prxs = subing.CesrSuber(db=self,
                                     subkey='prxs.',
                                     klas=coring.Cipher)
        self.nxts = subing.CesrSuber(db=self,
                                     subkey='nxts.',
                                     klas=coring.Cipher)
        self.mhabs = subing.CesrSuber(db=self,
                                      subkey='mhabs.',
                                      klas=coring.Prefixer)
        self.pres = koming.Komer(db=self,
                                 subkey='pres.',
                                 schema=Prefix, )  # New Prefix
        self.sprms = koming.Komer(db=self,
                                  subkey='sprms.',
                                  schema=SaltyPrm, )  # New Salty Parameters
        self.sits = koming.Komer(db=self,
                                 subkey='sits.',
                                 schema=PreSit, )  # Prefix Situation
        self.pubs = koming.Komer(db=self,
                                 subkey='pubs.',
                                 schema=PubSet, )  # public key set at pre.ridx
        return self.opened


class RemoteManager:
    def __init__(self, hby, rb: RemoteKeeper = None):
        self.hby = hby
        self.rb = rb if rb is not None else RemoteKeeper(name=hby.name,
                                                         base=hby.base,
                                                         temp=hby.temp,
                                                         reopen=True,
                                                         clear=False,
                                                         headDirPath=hby.db.headDirPath)

    def get(self, algo: Algos = None, pre=None):
        if pre is not None:
            if (pp := self.rb.pres.get(pre)) is None:
                raise ValueError("Attempt to load nonexistent pre={}.".format(pre))
            algo = pp.algo

        match algo:
            case Algos.salty:
                return SaltyManager(rb=self.rb)
            case Algos.randy:
                return RandyManager(rb=self.rb)
            case Algos.group:
                return GroupManager(rb=self.rb, rm=self)
            case _:
                return ExternKeeper(rb=self.rb)

    @property
    def sxlt(self):
        return self.rb.gbls.get("sxlt")

    @sxlt.setter
    def sxlt(self, sxlt):
        self.rb.gbls.pin("sxlt", sxlt)

    def delete_sxlt(self):
        return self.rb.gbls.rem("sxlt")


class SaltyManager:

    def __init__(self, rb: RemoteKeeper):
        self.rb = rb

    def incept(self, pre, *, icodes, ncodes, sxlt, dcode=MtrDex.Blake3_256, pidx=0, kidx=0, stem="", tier=Tiers.low,
               transferable=False):

        pp = Prefix(
            pidx=pidx,
            algo=Algos.salty
        )

        sp = SaltyPrm(sxlt=sxlt,
                      pidx=pidx,
                      kidx=kidx,
                      stem=stem,
                      tier=tier,
                      icodes=icodes,
                      ncodes=ncodes,
                      dcode=dcode,
                      transferable=transferable
                      )

        if not self.rb.pres.put(pre, val=pp):
            raise ValueError("Already incepted pre={}.".format(pre))

        if not self.rb.sprms.put(pre, val=sp):
            raise ValueError("Already incepted prm for pre={}.".format(pre))

    def rotate(self, pre, ncodes, pidx, kidx, stem, sxlt, icodes, tier, transferable, dcode=MtrDex.Blake3_256):
        if (pp := self.rb.pres.get(pre)) is None or pp.algo != Algos.salty:
            raise ValueError("Attempt to rotate nonexistent or invalid pre={}.".format(pre))

        if (sp := self.rb.sprms.get(pre)) is None:
            raise ValueError("Attempt to rotate nonexistent or invalid pre={}.".format(pre))

        sp = SaltyPrm(pidx=pidx,
                      sxlt=sxlt,
                      kidx=kidx,
                      stem=stem,
                      tier=sp.tier,
                      icodes=sp.icodes,
                      ncodes=ncodes,
                      dcode=dcode,
                      transferable=transferable
                      )

        if not self.rb.sprms.pin(pre, val=sp):
            raise ValueError("Unable to rotate salty prms for pre={}.".format(pre))

    def params(self, pre):
        if (pp := self.rb.pres.get(pre)) is None or pp.algo != Algos.salty:
            raise ValueError("Attempt to load nonexistent or invalid pre={}.".format(pre))

        if (pp := self.rb.sprms.get(pre)) is None:
            raise ValueError("Attempt to load nonexistent pre={}.".format(pre))

        prms = dict(
            salty=asdict(pp)
        )

        return prms


class RandyManager:

    def __init__(self, rb: RemoteKeeper):
        self.rb = rb

    def incept(self, pre, verfers, digers, prxs, nxts, transferable):

        pp = Prefix(
            algo=Algos.randy
        )

        if not self.rb.pres.put(pre, val=pp):
            raise ValueError("Already incepted pre={}.".format(pre))

        dt = helping.nowIso8601()
        ps = PreSit(
            new=PubLot(pubs=[verfer.qb64 for verfer in verfers],
                       dt=dt),
            nxt=PubLot(pubs=[diger.qb64 for diger in digers],
                       dt=dt))

        if not self.rb.sits.put(pre, val=ps):
            raise ValueError("Already incepted sit for pre={}.".format(pre))

        # Secret to encrypt here
        if len(prxs) != len(verfers):
            raise ValueError("If encrypted private keys are provided, must match verfers")

        for idx, prx in enumerate(prxs):
            cipher = coring.Cipher(qb64=prx)
            self.rb.prxs.put(keys=verfers[idx].qb64b, val=cipher)

        if nxts is not None:
            if len(nxts) != len(digers):
                raise ValueError("If encrypted private next keys are provided, must match digers")

            for idx, prx in enumerate(nxts):
                cipher = coring.Cipher(qb64=prx)
                self.rb.nxts.put(keys=digers[idx].qb64b, val=cipher)

    def rotate(self, pre, verfers, digers, prxs, nxts, transferable):
        if (pp := self.rb.pres.get(pre)) is None or pp.algo != Algos.randy:
            raise ValueError("Attempt to rotate non-existant or invalid pre={}.".format(pre))

        dt = helping.nowIso8601()
        ps = PreSit(
            new=PubLot(pubs=[verfer.qb64 for verfer in verfers],
                       dt=dt),
            nxt=PubLot(pubs=[diger.qb64 for diger in digers],
                       dt=dt))

        if not self.rb.sits.pin(pre, val=ps):
            raise ValueError("Already incepted sit for pre={}.".format(pre))

        # Secret to encrypt here
        if len(prxs) != len(verfers):
            raise ValueError("If encrypted private keys are provided, must match verfers")

        for idx, prx in enumerate(prxs):
            cipher = coring.Cipher(qb64=prx)
            self.rb.prxs.put(keys=verfers[idx].qb64b, val=cipher)

        if nxts is not None:
            if len(nxts) != len(digers):
                raise ValueError("If encrypted private keys are provided, must match verfers")

            for idx, prx in enumerate(nxts):
                cipher = coring.Cipher(qb64=prx)
                self.rb.nxts.put(keys=digers[idx].qb64b, val=cipher)

    def params(self, pre):
        if (pp := self.rb.pres.get(pre)) is None or pp.algo != Algos.randy:
            raise ValueError("Attempt to load nonexistent or invalid pre={}.".format(pre))

        prxs = []
        if (ps := self.rb.sits.get(pre)) is None:
            raise ValueError("Attempt to load nonexistent pre={}.".format(pre))

        for pub in ps.new.pubs:
            if (prx := self.rb.prxs.get(keys=pub)) is not None:
                prxs.append(prx)

        nxts = []
        for pub in ps.nxt.pubs:
            if (nxt := self.rb.nxts.get(keys=pub)) is not None:
                nxts.append(nxt)

        prms = dict(
            randy=dict(
                prxs=[prx.qb64 for prx in prxs],
                nxts=[nxt.qb64 for nxt in nxts],
            ),
        )

        return prms


class GroupManager:

    def __init__(self, rb: RemoteKeeper, rm: RemoteManager):
        self.rb = rb
        self.rm = rm

    def incept(self, pre, mpre, verfers, digers):
        pp = Prefix(
            algo=Algos.group
        )

        if not self.rb.pres.put(pre, val=pp):
            raise ValueError("Already incepted pre={}.".format(pre))

        if not self.rb.mhabs.put(pre, val=coring.Prefixer(qb64=mpre)):
            raise ValueError("Already incepted pre={}.".format(pre))

        dt = helping.nowIso8601()
        ps = PreSit(
            new=PubLot(pubs=[verfer.qb64 for verfer in verfers],
                       dt=dt),
            nxt=PubLot(pubs=[diger.qb64 for diger in digers],
                       dt=dt))

        if not self.rb.sits.put(pre, val=ps):
            raise ValueError("Already incepted sit for pre={}.".format(pre))

    def rotate(self, pre, verfers, digers):
        if (pp := self.rb.pres.get(pre)) is None or pp.algo != Algos.group:
            raise ValueError(f"Attempt to rotate nonexistant or invalid  pre={pre}, algo={pp.algo}.")

        dt = helping.nowIso8601()
        ps = PreSit(
            new=PubLot(pubs=[verfer.qb64 for verfer in verfers],
                       dt=dt),
            nxt=PubLot(pubs=[diger.qb64 for diger in digers],
                       dt=dt))

        if not self.rb.sits.pin(pre, val=ps):
            raise ValueError(f"Error saving sit rotating pre={pre}.")

    def params(self, pre):
        if (pp := self.rb.pres.get(pre)) is None or pp.algo != Algos.group:
            raise ValueError(f"Attempt to load nonexistent or invalid pre={pre}.")

        if (mpre := self.rb.mhabs.get(pre)) is None:
            raise ValueError("Attempt to load nonexistent pre={}.".format(pre))

        if (ps := self.rb.sits.get(pre)) is None:
            raise ValueError(f"Attempt to load invalid sit for pre={pre}.")

        prms = dict(
            group=dict(
                mhab=self.rm.get(pre=mpre.qb64).params(mpre.qb64),
                keys=[verfer for verfer in ps.new.pubs],
                ndigs=[diger for diger in ps.nxt.pubs]
            )
        )

        return prms


class ExternKeeper:

    def __init__(self, rb: RemoteKeeper):
        self.rb = rb

    def incept(self, **kwargs):
        pass


# -*- encoding: utf-8 -*-
"""
KERIA
keria.app.aiding module

"""
import json
from dataclasses import asdict
from urllib.parse import urlparse, urljoin

import falcon
from keri import kering
from keri.app import habbing
from keri.app.keeping import Algos
from keri.core import coring, serdering
from keri.core.coring import Ilks
from keri.db import dbing
from keri.help import ogler
from mnemonic import mnemonic

from ..core import longrunning, httping

logger = ogler.getLogger()


def loadEnds(app, agency, authn):
    groupEnd = AgentResourceEnd(agency=agency, authn=authn)
    app.add_route("/agent/{caid}", groupEnd)

    aidsEnd = IdentifierCollectionEnd()
    app.add_route("/identifiers", aidsEnd)
    submitEnd = IdentifierSubmitResourceEnd()    
    app.add_route("/identifiers/{name}/submit", submitEnd)    
    aidEnd = IdentifierResourceEnd()
    app.add_route("/identifiers/{name}", aidEnd)

    aidOOBIsEnd = IdentifierOOBICollectionEnd()
    app.add_route("/identifiers/{name}/oobis", aidOOBIsEnd)

    endRolesEnd = EndRoleCollectionEnd()
    app.add_route("/identifiers/{name}/endroles", endRolesEnd)
    app.add_route("/identifiers/{name}/endroles/{role}", endRolesEnd)
    app.add_route("/endroles/{aid}", endRolesEnd)
    app.add_route("/endroles/{aid}/{role}", endRolesEnd)

    endRoleEnd = EndRoleResourceEnd()
    app.add_route("/identifiers/{name}/endroles/{role}/{eid}", endRoleEnd)

    rpyEscrowEnd = RpyEscrowCollectionEnd()
    app.add_route("/escrows/rpy", rpyEscrowEnd)

    chaEnd = ChallengeCollectionEnd()
    app.add_route("/challenges", chaEnd)
    chaResEnd = ChallengeResourceEnd()
    app.add_route("/challenges/{name}", chaResEnd)
    chaVerResEnd = ChallengeVerifyResourceEnd()
    app.add_route("/challenges_verify/{source}", chaVerResEnd)

    contactColEnd = ContactCollectionEnd()
    app.add_route("/contacts", contactColEnd)
    contactResEnd = ContactResourceEnd()
    app.add_route("/contacts/{prefix}", contactResEnd)
    contactImgEnd = ContactImageResourceEnd()
    app.add_route("/contacts/{prefix}/img", contactImgEnd)

    groupEnd = GroupMemberCollectionEnd()
    app.add_route("/identifiers/{name}/members", groupEnd)

    return aidEnd


class AgentResourceEnd:
    """ Resource class for getting agent specific launch information """

    def __init__(self, agency, authn):
        self.agency = agency
        self.authn = authn

    def on_get(self, _, rep, caid):
        """ GET endpoint for Keystores

        Get keystore status

        Args:
            _: falcon.Request HTTP request
            rep: falcon.Response HTTP response
            caid(str): qb64 identifier prefix of Controller

        """
        agent = self.agency.get(caid)
        if agent is None:
            raise falcon.HTTPNotFound(description=f"not agent found for controller {caid}")

        if agent.pre not in agent.hby.kevers:
            raise falcon.HTTPBadRequest(description=f"invalid agent configuration, {agent.pre} not found")

        if agent.caid not in agent.hby.kevers:
            raise falcon.HTTPBadRequest(description=f"invalid controller configuration, {agent.caid} not found")

        pidx = agent.hby.db.habs.cntAll()

        state = asdict(agent.hby.kevers[agent.caid].state())
        key = dbing.dgKey(state['i'], state['ee']['d'])  # digest key
        msg = agent.hby.db.getEvt(key)
        eserder = serdering.SerderKERI(raw=bytes(msg))

        body = dict(
            agent=asdict(agent.hby.kevers[agent.pre].state()),
            controller=dict(
                state=state,
                ee=eserder.ked
            ),
            pidx=pidx
        )

        if (sxlt := agent.mgr.sxlt) is not None:
            body["sxlt"] = sxlt

        rep.content_type = "application/json"
        rep.data = json.dumps(body).encode("utf-8")
        rep.status = falcon.HTTP_200

    def on_put(self, req, rep, caid):
        """

        Parameters:
            req (Request): falcon.Request HTTP request
            rep (Response): falcon.Response HTTP response
            caid(str): qb64 identifier prefix of Controller

        """
        agent = self.agency.get(caid)
        if agent is None:
            raise falcon.HTTPNotFound(description=f"no agent for {caid}")

        typ = req.params.get("type")
        if typ == "ixn":
            ixn = self.interact(req, rep, agent, caid)
            self.anchorSeals(agent, ixn)
            return

        body = req.get_media()

        if "rot" not in body:
            raise falcon.HTTPBadRequest(description="required field 'rot' missing from body")

        if "sigs" not in body:
            raise falcon.HTTPBadRequest(description="required field 'sigs' missing from body")

        if "sxlt" not in body:
            raise falcon.HTTPBadRequest(description="required field 'sxlt' missing from body")

        if "keys" not in body:
            raise falcon.HTTPBadRequest(description="required field 'keys' missing from body")

        rot = serdering.SerderKERI(sad=body["rot"])
        sigs = body["sigs"]

        ctrlHab = agent.hby.habByName(caid, ns="agent")
        ctrlHab.rotate(serder=rot, sigers=[coring.Siger(qb64=sig) for sig in sigs])

        if not self.authn.verify(req):
            raise falcon.HTTPForbidden(description="invalid signature on request")
        
        sxlt = body["sxlt"]
        agent.mgr.sxlt = sxlt

        keys = body["keys"]
        for pre, val in keys.items():
            if "sxlt" in val:
                if (sp := agent.mgr.rb.sprms.get(pre)) is None:
                    raise ValueError("Attempt to update sxlt for nonexistent or invalid pre={}.".format(pre))

                sp.sxlt = val["sxlt"]

                if not agent.mgr.rb.sprms.pin(pre, val=sp):
                    raise ValueError("Unable to update sxlt prms for pre={}.".format(pre))

            elif "prxs" in val:
                hab = agent.hby.habs[pre]
                verfers = hab.kever.verfers
                digers = hab.kever.digers
                prxs = val["prxs"]

                for idx, prx in enumerate(prxs):
                    cipher = coring.Cipher(qb64=prx)
                    agent.mgr.rb.prxs.put(keys=verfers[idx].qb64b, val=cipher)

                if "nxts" in val:
                    nxts = val["nxts"]
                    if len(nxts) != len(digers):
                        raise ValueError("If encrypted private next keys are provided, must match digers")

                    for idx, prx in enumerate(nxts):
                        cipher = coring.Cipher(qb64=prx)
                        agent.mgr.rb.nxts.put(keys=digers[idx].qb64b, val=cipher)

        agent.mgr.delete_sxlt()

        rep.status = falcon.HTTP_204

    @staticmethod
    def interact(req, rep, agent, caid):
        body = req.get_media()

        if "ixn" not in body:
            raise falcon.HTTPBadRequest(description="required field 'ixn' missing from body")

        if "sigs" not in body:
            raise falcon.HTTPBadRequest(description="required field 'sigs' missing from body")

        ked = body['ixn']
        sigs = body['sigs']
        ixn = serdering.SerderKERI(sad=ked)
        sigers = [coring.Siger(qb64=sig) for sig in sigs]

        ctrlHab = agent.hby.habByName(caid, ns="agent")

        ctrlHab.interact(serder=ixn, sigers=sigers)
        agent.agentHab.kvy.processEvent(serder=ixn, sigers=sigers)

        rep.status = falcon.HTTP_204

        return ixn

    @staticmethod
    def anchorSeals(agent, ixn):
        a = ixn.ked["a"]
        if len(a) == 0:
            return

        delegator = coring.Saider(qb64=ixn.ked['d'])
        delegatorsn = coring.Seqner(snh=ixn.ked['s'])

        seal = a[0]
        prefixer = coring.Prefixer(qb64=seal['i'])
        saider = coring.Saider(qb64=seal["d"])

        couple = delegatorsn.qb64b + delegator.qb64b
        dgkey = dbing.dgKey(prefixer.qb64b, saider.qb64)
        agent.hby.db.setAes(dgkey, couple)  # authorizer event seal (delegator/issuer)


class IdentifierCollectionEnd:
    """ Resource class for creating and managing identifiers """

    @staticmethod
    def on_options(req, rep):
        rep.add_header("Accept-Ranges", "aids")
        rep.status = falcon.HTTP_200

    @staticmethod
    def on_get(req, rep):
        """ Identifier List GET endpoint

        Parameters:
            req: falcon.Request HTTP request
            rep: falcon.Response HTTP response

        """
        agent = req.context.agent
        res = []

        rng = req.get_header("Range")
        if rng is None:
            rep.status = falcon.HTTP_200
            start = 0
            end = 9
        else:
            rep.status = falcon.HTTP_206
            start, end = httping.parseRangeHeader(rng, "aids")

        count = agent.hby.db.habs.cntAll()
        it = agent.hby.db.habs.getItemIter()
        for _ in range(start):
            try:
                next(it)
            except StopIteration:
                break

        for name, habord in it:
            name = ".".join(name)  # detupleize the database key name

            hab = agent.hby.habByName(name)
            data = info(hab, agent.mgr)
            res.append(data)

            if (not end == -1) and len(res) == (end - start) + 1:
                break

        end = start + (len(res) - 1) if len(res) > 0 else 0
        rep.set_header("Accept-Ranges", "aids")
        rep.set_header("Content-Range", f"aids {start}-{end}/{count}")
        rep.content_type = "application/json"
        rep.data = json.dumps(res).encode("utf-8")

    @staticmethod
    def on_post(req, rep):
        """ Inception event POST endpoint

        Parameters:
            req (Request): falcon.Request HTTP request object
            rep (Response): falcon.Response HTTP response object

        """
        agent = req.context.agent
        try:
            body = req.get_media()
            icp = httping.getRequiredParam(body, "icp")
            name = httping.getRequiredParam(body, "name")
            sigs = httping.getRequiredParam(body, "sigs")

            serder = serdering.SerderKERI(sad=icp)

            sigers = [coring.Siger(qb64=sig) for sig in sigs]

            if agent.hby.habByName(name) is not None:
                raise falcon.HTTPBadRequest(title=f"AID with name {name} already incepted")

            if 'b' in icp:
                for wit in icp['b']:
                    urls = agent.agentHab.fetchUrls(eid=wit, scheme=kering.Schemes.http)
                    if not urls and wit not in agent.hby.kevers:
                        raise falcon.HTTPBadRequest(description=f'unknown witness {wit}')

            if 'di' in icp and icp["di"] not in agent.hby.kevers:
                raise falcon.HTTPBadRequest(description=f'unknown delegator {icp["di"]}')

            # client is requesting agent to join multisig group
            if "group" in body:
                group = body["group"]

                if "mhab" not in group:
                    raise falcon.HTTPBadRequest(description=f'required field "mhab" missing from body.group')
                mpre = group["mhab"]["prefix"]

                if mpre not in agent.hby.habs:
                    raise falcon.HTTPBadRequest(description=f'signing member {mpre} not a local AID')
                mhab = agent.hby.habs[mpre]

                if "keys" not in group:
                    raise falcon.HTTPBadRequest(description=f'required field "keys" missing from body.group')
                keys = group["keys"]
                verfers = [coring.Verfer(qb64=key) for key in keys]

                if mhab.kever.fetchLatestContribTo(verfers=verfers) is None:
                    raise falcon.HTTPBadRequest(description=f"Member hab={mhab.pre} not a participant in "
                                                            f"event for this group hab.")

                if "ndigs" not in group:
                    raise falcon.HTTPBadRequest(description=f'required field "ndigs" missing from body.group')
                ndigs = group["ndigs"]
                digers = [coring.Diger(qb64=ndig) for ndig in ndigs]

                smids = httping.getRequiredParam(body, "smids")
                rmids = httping.getRequiredParam(body, "rmids")

                hab = agent.hby.makeSignifyGroupHab(name, mhab=mhab, serder=serder, sigers=sigers)
                try:
                    agent.inceptGroup(pre=serder.pre, mpre=mhab.pre, verfers=verfers, digers=digers)
                except ValueError as e:
                    agent.hby.deleteHab(name=name)
                    raise falcon.HTTPInternalServerError(description=f"{e.args[0]}")

                # Generate response, a long running operaton indicator for the type
                agent.groups.append(dict(pre=hab.pre, serder=serder, sigers=sigers, smids=smids, rmids=rmids))
                op = agent.monitor.submit(serder.pre, longrunning.OpTypes.group, metadata=dict(sn=0))

                rep.content_type = "application/json"
                rep.status = falcon.HTTP_202
                rep.data = op.to_json().encode("utf-8")

            else:
                # client is requesting that the Agent track the Salty parameters
                if Algos.salty in body:
                    salt = body[Algos.salty]
                    hab = agent.hby.makeSignifyHab(name, serder=serder, sigers=sigers)
                    try:
                        agent.inceptSalty(pre=serder.pre, **salt)
                    except ValueError as e:
                        agent.hby.deleteHab(name=name)
                        raise falcon.HTTPInternalServerError(description=f"{e.args[0]}")

                # client is storing encrypted randomly generated key material on agent
                elif Algos.randy in body:
                    rand = body[Algos.randy]
                    hab = agent.hby.makeSignifyHab(name, serder=serder, sigers=sigers)
                    try:
                        agent.inceptRandy(pre=serder.pre, verfers=serder.verfers, digers=serder.ndigers, **rand)
                    except ValueError as e:
                        agent.hby.deleteHab(name=name)
                        raise falcon.HTTPInternalServerError(description=f"{e.args[0]}")

                elif Algos.extern in body:
                    extern = body[Algos.extern]
                    hab = agent.hby.makeSignifyHab(name, serder=serder, sigers=sigers)
                    try:
                        agent.inceptExtern(pre=serder.pre, verfers=serder.verfers, digers=serder.ndigers, **extern)
                    except ValueError as e:
                        agent.hby.deleteHab(name=name)
                        raise falcon.HTTPInternalServerError(description=f"{e.args[0]}")

                else:
                    raise falcon.HTTPBadRequest(
                        description="invalid request: one of group, rand or salt field required")

                # create Hab and incept the key store (if any)
                # Generate response, either the serder or a long running operaton indicator for the type
                rep.content_type = "application/json"
                if hab.kever.delegator:
                    agent.anchors.append(dict(pre=hab.pre, sn=0))
                    op = agent.monitor.submit(hab.kever.prefixer.qb64, longrunning.OpTypes.delegation,
                                              metadata=dict(pre=hab.pre, sn=0))
                    rep.status = falcon.HTTP_202
                    rep.data = op.to_json().encode("utf-8")

                elif hab.kever.wits:
                    agent.witners.append(dict(serder=serder))
                    op = agent.monitor.submit(hab.kever.prefixer.qb64, longrunning.OpTypes.witness,
                                              metadata=dict(sn=0))
                    rep.status = falcon.HTTP_202
                    rep.data = op.to_json().encode("utf-8")

                else:
                    rep.status = falcon.HTTP_202
                    op = agent.monitor.submit(hab.kever.prefixer.qb64, longrunning.OpTypes.done,
                                              metadata=dict(response=serder.ked))
                    rep.data = op.to_json().encode("utf-8")

        except (kering.AuthError, ValueError) as e:
            raise falcon.HTTPBadRequest(description=e.args[0])


class IdentifierResourceEnd:
    """ Resource class for updating and deleting identifiers """

    @staticmethod
    def on_get(req, rep, name):
        """ Identifier GET endpoint

        Parameters:
            req: falcon.Request HTTP request
            rep: falcon.Response HTTP response
            name (str): human readable name for Hab to GET

        """
        if not name:
            raise falcon.HTTPBadRequest(description="name is required")

        agent = req.context.agent
        hab = agent.hby.habByName(name)
        if hab is None:
            raise falcon.HTTPNotFound(description=f"{name} is not a valid identifier name")

        data = info(hab, agent.mgr, full=True)
        rep.status = falcon.HTTP_200
        rep.content_type = "application/json"
        rep.data = json.dumps(data).encode("utf-8")

    def on_put(self, req, rep, name):
        """ Identifier UPDATE endpoint

        Parameters:
            req (Request): falcon.Request HTTP request object
            rep (Response): falcon.Response HTTP response object
            name (str): human readable name for Hab to rotate or interact

        """
        agent = req.context.agent
        try:
            body = req.get_media()
            typ = Ilks.ixn if req.params.get("type") == "ixn" else Ilks.rot

            if typ in (Ilks.rot,):
                op = self.rotate(agent, name, body)
            else:
                op = self.interact(agent, name, body)

            rep.status = falcon.HTTP_200
            rep.content_type = "application/json"
            rep.data = op.to_json().encode("utf-8")

        except (kering.AuthError, ValueError) as e:
            raise falcon.HTTPBadRequest(description=e.args[0])

    @staticmethod
    def rotate(agent, name, body):
        hab = agent.hby.habByName(name)
        if hab is None:
            raise falcon.HTTPNotFound(title=f"No AID with name {name} found")
               
        rot = body.get("rot")
        if rot is None:
            raise falcon.HTTPBadRequest(title="invalid rotation",
                                        description=f"required field 'rot' missing from request")

        if 'ba' in rot:
            for wit in rot['ba']:
                urls = agent.agentHab.fetchUrls(eid=wit, scheme=kering.Schemes.http)
                if not urls and wit not in agent.hby.kevers:
                    raise falcon.HTTPBadRequest(description=f'unknown witness {wit}')

        sigs = body.get("sigs")
        if sigs is None or len(sigs) == 0:
            raise falcon.HTTPBadRequest(title="invalid rotation",
                                        description=f"required field 'sigs' missing from request")

        serder = serdering.SerderKERI(sad=rot)
        sigers = [coring.Siger(qb64=sig) for sig in sigs]

        hab.rotate(serder=serder, sigers=sigers)

        if Algos.salty in body:
            salt = body[Algos.salty]
            keeper = agent.mgr.get(Algos.salty)

            try:
                keeper.rotate(pre=serder.pre, **salt)
            except ValueError as e:
                agent.hby.deleteHab(name=name)
                raise falcon.HTTPInternalServerError(description=f"{e.args[0]}")

        elif Algos.randy in body:
            rand = body[Algos.randy]
            keeper = agent.mgr.get(Algos.randy)

            keeper.rotate(pre=serder.pre, verfers=serder.verfers, digers=serder.ndigers, **rand)

        elif Algos.group in body:
            keeper = agent.mgr.get(Algos.group)

            keeper.rotate(pre=serder.pre, verfers=serder.verfers, digers=serder.ndigers)

            smids = httping.getRequiredParam(body, "smids")
            rmids = httping.getRequiredParam(body, "rmids")

            agent.groups.append(dict(pre=hab.pre, serder=serder, sigers=sigers, smids=smids, rmids=rmids))
            op = agent.monitor.submit(serder.pre, longrunning.OpTypes.group, metadata=dict(sn=serder.sn))

            return op

        if hab.kever.delegator:
            agent.anchors.append(dict(alias=name, pre=hab.pre, sn=serder.sn))
            op = agent.monitor.submit(hab.kever.prefixer.qb64, longrunning.OpTypes.delegation,
                                      metadata=dict(pre=hab.pre, sn=serder.sn))
            return op

        if hab.kever.wits:
            agent.witners.append(dict(serder=serder))
            op = agent.monitor.submit(hab.kever.prefixer.qb64, longrunning.OpTypes.witness,
                                      metadata=dict(sn=serder.sn))
            return op

        op = agent.monitor.submit(hab.kever.prefixer.qb64, longrunning.OpTypes.done,
                                  metadata=dict(response=serder.ked))
        return op

    @staticmethod
    def interact(agent, name, body):
        hab = agent.hby.habByName(name)
        if hab is None:
            raise falcon.HTTPNotFound(title=f"No AID {name} found")

        ixn = body.get("ixn")
        if ixn is None:
            raise falcon.HTTPBadRequest(title="invalid interaction",
                                        description=f"required field 'ixn' missing from request")

        sigs = body.get("sigs")
        if sigs is None or len(sigs) == 0:
            raise falcon.HTTPBadRequest(title="invalid interaction",
                                        description=f"required field 'sigs' missing from request")

        serder = serdering.SerderKERI(sad=ixn)
        sigers = [coring.Siger(qb64=sig) for sig in sigs]

        hab.interact(serder=serder, sigers=sigers)

        if "group" in body:
            agent.groups.append(dict(pre=hab.pre, serder=serder, sigers=sigers))
            op = agent.monitor.submit(serder.pre, longrunning.OpTypes.group, metadata=dict(sn=serder.sn))

            return op

        if hab.kever.wits:
            agent.witners.append(dict(serder=serder))
            op = agent.monitor.submit(hab.kever.prefixer.qb64, longrunning.OpTypes.witness,
                                      metadata=dict(sn=serder.sn))
            return op

        op = agent.monitor.submit(hab.kever.prefixer.qb64, longrunning.OpTypes.done,
                                  metadata=dict(response=serder.ked))
        return op


def info(hab, rm, full=False):
    data = dict(
        name=hab.name,
        prefix=hab.pre,
    )

    if not isinstance(hab, habbing.SignifyHab) and not isinstance(hab, habbing.SignifyGroupHab):
        raise kering.ConfigurationError(f"agent only allows SignifyHab instances, {type(hab)}")

    keeper = rm.get(pre=hab.pre)
    data.update(keeper.params(pre=hab.pre))
    if isinstance(hab, habbing.SignifyGroupHab):
        data["group"]["mhab"] = info(hab.mhab, rm, full)

    if hab.accepted and full:
        kever = hab.kevers[hab.pre]
        data["transferable"] = kever.transferable
        data["state"] = asdict(kever.state())
        dgkey = dbing.dgKey(kever.prefixer.qb64b, kever.serder.saidb)
        wigs = hab.db.getWigs(dgkey)
        data["windexes"] = [coring.Siger(qb64b=bytes(wig)).index for wig in wigs]

    return data

class IdentifierSubmitResourceEnd:
    """ Resource class to submit identifier to its witnesses for receipts """

    @staticmethod
    def on_put(req, rep, name):
        """ Submit request POST endpoint

        Parameters:
            req (Request): falcon.Request HTTP request object
            rep (Response): falcon.Response HTTP response object
            name (str): human readable name for Hab to submit

        """
        agent = req.context.agent
        try:
            hab = agent.hby.habByName(name)
            if not hab:
                raise falcon.HTTPNotFound(description=f"invalid alias {name}")

            oid = hab.pre
            op = agent.monitor.submit(oid, longrunning.OpTypes.submit, metadata=dict(pre=hab.pre))

            rep.content_type = "application/json"
            rep.status = falcon.HTTP_202
            rep.data = op.to_json().encode("utf-8")

        except (kering.AuthError, ValueError) as e:
            raise falcon.HTTPBadRequest(description=e.args[0])     
               
class IdentifierOOBICollectionEnd:
    """
      This class represents the OOBI subresource collection endpoint for identifiers

    """

    @staticmethod
    def on_get(req, rep, name):
        """ Identifier GET endpoint

        Parameters:
            req: falcon.Request HTTP request
            rep: falcon.Response HTTP response
            name (str): human readable name for Hab to GET

        """
        agent = req.context.agent
        if not name:
            raise falcon.HTTPBadRequest(description="name is required")

        hab = agent.hby.habByName(name)
        if not hab:
            raise falcon.HTTPNotFound(description="invalid alias {name}")

        if "role" not in req.params:
            raise falcon.HTTPBadRequest(description="role parameter required")

        role = req.params["role"]

        res = dict(role=role)
        if role in (kering.Roles.witness,):  # Fetch URL OOBIs for all witnesses
            oobis = []
            for wit in hab.kever.wits:
                urls = hab.fetchUrls(eid=wit, scheme=kering.Schemes.http) or hab.fetchUrls(eid=wit, scheme=kering.Schemes.https)
                if not urls:
                    raise falcon.HTTPNotFound(description=f"unable to query witness {wit}, no http endpoint")

                url = urls[kering.Schemes.http] if kering.Schemes.http in urls else urls[kering.Schemes.https]
                up = urlparse(url)
                oobis.append(urljoin(up.geturl(), f"/oobi/{hab.pre}/witness/{wit}"))
            res["oobis"] = oobis
        elif role in (kering.Roles.controller,):  # Fetch any controller URL OOBIs
            oobis = []
            urls = hab.fetchUrls(eid=hab.pre, scheme=kering.Schemes.http) or hab.fetchUrls(eid=hab.pre, scheme=kering.Schemes.https)
            if not urls:
                raise falcon.HTTPNotFound(description=f"unable to query controller {hab.pre}, no http endpoint")

            url = urls[kering.Schemes.http] if kering.Schemes.http in urls else urls[kering.Schemes.https]
            up = urlparse(url)
            oobis.append(urljoin(up.geturl(), f"/oobi/{hab.pre}/controller"))
            res["oobis"] = oobis
        elif role in (kering.Roles.agent,):  # Fetch URL OOBIs for all witnesses
            roleUrls = hab.fetchRoleUrls(cid=hab.pre, role=kering.Roles.agent, scheme=kering.Schemes.http) or hab.fetchRoleUrls(cid=hab.pre, role=kering.Roles.agent, scheme=kering.Schemes.https)
            if kering.Roles.agent not in roleUrls:
                res['oobis'] = []
            else:
                aoobis = roleUrls[kering.Roles.agent]

                oobis = list()
                for agent in set(aoobis.keys()):
                    murls = aoobis.naball(agent)
                    for murl in murls:
                        urls = []
                        if kering.Schemes.http in murl:
                            urls.extend(murl.naball(kering.Schemes.http))
                        if kering.Schemes.https in murl:
                            urls.extend(murl.naball(kering.Schemes.https))
                        for url in urls:
                            up = urlparse(url)
                            oobis.append(urljoin(up.geturl(), f"/oobi/{hab.pre}/agent/{agent}"))

                res["oobis"] = oobis
        elif role in (kering.Roles.mailbox,):  # Fetch URL OOBIs for all witnesses
            roleUrls = (hab.fetchRoleUrls(cid=hab.pre, role=kering.Roles.mailbox, scheme=kering.Schemes.http) or
                        hab.fetchRoleUrls(cid=hab.pre, role=kering.Roles.mailbox, scheme=kering.Schemes.https))
            if kering.Roles.mailbox not in roleUrls:
                res['oobis'] = []
            else:
                aoobis = roleUrls[kering.Roles.mailbox]

                oobis = list()
                for mailbox in set(aoobis.keys()):
                    murls = aoobis.naball(mailbox)
                    for murl in murls:
                        urls = []
                        if kering.Schemes.http in murl:
                            urls.extend(murl.naball(kering.Schemes.http))
                        if kering.Schemes.https in murl:
                            urls.extend(murl.naball(kering.Schemes.https))
                        for url in urls:
                            up = urlparse(url)
                            oobis.append(urljoin(up.geturl(), f"/oobi/{hab.pre}/mailbox/{mailbox}"))

                res["oobis"] = oobis
        else:
            raise falcon.HTTPBadRequest(description=f"unsupport role type {role} for oobi request")

        rep.status = falcon.HTTP_200
        rep.content_type = "application/json"
        rep.data = json.dumps(res).encode("utf-8")


class EndRoleCollectionEnd:

    @staticmethod
    def on_get(req, rep, name=None, aid=None, role=None):
        """  GET endpoint for end role collection

        Parameters:
            req (Request): falcon HTTP request object
            rep (Response): falcon HTTP response object
            name (str): human readable alias for AID
            aid (str): aid to use instead of name
            role (str): optional role to search for

        """
        agent = req.context.agent

        if name is not None:
            hab = agent.hby.habByName(name)
            if hab is None:
                raise falcon.errors.HTTPNotFound(description=f"invalid alias {name}")
            pre = hab.pre
        elif aid is not None:
            pre = aid
        else:
            raise falcon.HTTPBadRequest(description="either `aid` or `name` are required in the path")

        if role is not None:
            keys = (pre, role,)
        else:
            keys = (pre,)

        ends = []
        for (_, erole, eid), end in agent.hby.db.ends.getItemIter(keys=keys):
            ends.append(dict(cid=pre, role=erole, eid=eid))

        rep.status = falcon.HTTP_200
        rep.content_type = "application/json"
        rep.data = json.dumps(ends).encode("utf-8")

    @staticmethod
    def on_post(req, rep, name, aid=None, role=None):
        """ POST endpoint for end role collection

        Args:
            req (Request): Falcon HTTP request object
            rep (Response): Falcon HTTP response object
            name (str): human readable alias for AID
            aid (str): Not supported for POST.  If provided, a 404 is returned
            role (str): Not supported for POST.  If provided, a 404 is returned

        """
        if role is not None or aid is not None:
            raise falcon.HTTPNotFound(description="route not found")

        agent = req.context.agent
        body = req.get_media()

        rpy = httping.getRequiredParam(body, "rpy")
        rsigs = httping.getRequiredParam(body, "sigs")

        rserder = serdering.SerderKERI(sad=rpy)
        data = rserder.ked['a']
        pre = data['cid']
        role = data['role']
        eid = data['eid']

        hab = agent.hby.habByName(name)
        if hab is None:
            raise falcon.errors.HTTPNotFound(description=f"invalid alias {name}")

        if pre != hab.pre:
            raise falcon.errors.HTTPBadRequest(
                description=f"error trying to create end role for unknown local AID {pre}")

        rsigers = [coring.Siger(qb64=rsig) for rsig in rsigs]
        tsg = (hab.kever.prefixer, coring.Seqner(sn=hab.kever.sn), coring.Saider(qb64=hab.kever.serder.said), rsigers)
        try:
            agent.hby.rvy.processReply(rserder, tsgs=[tsg])
        except kering.UnverifiedReplyError:
            pass

        oid = ".".join([pre, role, eid])
        op = agent.monitor.submit(oid, longrunning.OpTypes.endrole, metadata=dict(cid=pre, role=role, eid=eid))

        rep.content_type = "application/json"
        rep.status = falcon.HTTP_202
        rep.data = op.to_json().encode("utf-8")


class EndRoleResourceEnd:

    def on_delete(self, req, rep):
        pass


class RpyEscrowCollectionEnd:

    @staticmethod
    def on_get(req, rep):
        agent = req.context.agent

        # Optional Route parameter
        route = req.params.get("route")
        keys = (route,) if route is not None else ()
        events = []
        for saider in agent.hby.db.rpes.get(keys=keys):
            serder = agent.hby.db.rpys.get(keys=(saider.qb64,))
            events.append(serder.ked)

        rep.set_header("Content-Type", "application/json")
        rep.status = falcon.HTTP_200
        rep.data = json.dumps(events).encode("utf-8")


class ChallengeCollectionEnd:
    """ Resource for Challenge/Response Endpoints """

    @staticmethod
    def on_get(req, rep):
        """ Challenge GET endpoint

        Parameters:
            req: falcon.Request HTTP request
            rep: falcon.Response HTTP response

        ---
        summary:  Get random list of words for a 2 factor auth challenge
        description:  Get the list of identifiers associated with this agent
        tags:
           - Challenge/Response
        parameters:
           - in: query
             name: strength
             schema:
                type: int
             description:  cryptographic strength of word list
             required: false
        responses:
            200:
              description: An array of random words
              content:
                  application/json:
                    schema:
                        description: Random word list
                        type: object
                        properties:
                            words:
                                type: array
                                description: random challenge word list
                                items:
                                    type: string

        """
        mnem = mnemonic.Mnemonic(language='english')
        s = req.params.get("strength")
        strength = int(s) if s is not None else 128

        words = mnem.generate(strength=strength)
        rep.status = falcon.HTTP_200
        rep.content_type = "application/json"
        msg = dict(words=words.split(" "))
        rep.data = json.dumps(msg).encode("utf-8")


class ChallengeResourceEnd:
    """ Resource for Challenge/Response Endpoints """

    @staticmethod
    def on_post(req, rep, name):
        """ Challenge POST endpoint

        Parameters:
            req: falcon.Request HTTP request
            rep: falcon.Response HTTP response
            name: human readable name of identifier to use to sign the challenge/response

        ---
        summary:  Sign challenge message and forward to peer identifier
        description:  Sign a challenge word list received out of bands and send `exn` peer to peer message
                      to recipient
        tags:
           - Challenge/Response
        parameters:
          - in: path
            name: name
            schema:
              type: string
            required: true
            description: Human readable alias for the identifier to create
        requestBody:
            required: true
            content:
              application/json:
                schema:
                    description: Challenge response
                    properties:
                        recipient:
                          type: string
                          description: human readable alias recipient identifier to send signed challenge to
                        words:
                          type: array
                          description:  challenge in form of word list
                          items:
                              type: string
        responses:
           202:
              description: Success submission of signed challenge/response
        """
        agent = req.context.agent
        hab = agent.hby.habByName(name)
        if hab is None:
            raise falcon.HTTPBadRequest(description="no matching Hab for alias {name}")

        body = req.get_media()
        if "exn" not in body or "sig" not in body or "recipient" not in body:
            raise falcon.HTTPBadRequest(description="challenge response requires 'words', 'sig' and 'recipient'")

        exn = body["exn"]
        sig = body["sig"]
        recpt = body["recipient"]
        serder = serdering.SerderKERI(sad=exn)

        ims = bytearray(serder.raw)
        ims.extend(sig.encode("utf-8"))

        agent.hby.psr.parseOne(ims=bytearray(ims))

        agent.exchanges.append(dict(said=serder.said, pre=hab.pre, rec=recpt, topic='challenge'))

        rep.status = falcon.HTTP_202


class ChallengeVerifyResourceEnd:
    """ Resource for Challenge/Response Verification Endpoints """

    @staticmethod
    def on_post(req, rep, source):
        """ Challenge POST endpoint

        Parameters:
            req: falcon.Request HTTP request
            rep: falcon.Response HTTP response
            source: qb64 AID of of source of signed response to verify

        ---
        summary:  Sign challenge message and forward to peer identifier
        description:  Sign a challenge word list received out of bands and send `exn` peer to peer message
                      to recipient
        tags:
           - Challenge/Response
        parameters:
          - in: path
            name: name
            schema:
              type: string
            required: true
            description: Human readable alias for the identifier to create
        requestBody:
            required: true
            content:
              application/json:
                schema:
                    description: Challenge response
                    properties:
                        recipient:
                          type: string
                          description: human readable alias recipient identifier to send signed challenge to
                        words:
                          type: array
                          description:  challenge in form of word list
                          items:
                              type: string
        responses:
           202:
              description: Success submission of signed challenge/response
        """
        agent = req.context.agent

        body = req.get_media()
        words = httping.getRequiredParam(body, "words")
        if source not in agent.hby.kevers:
            raise falcon.HTTPNotFound(description=f"challenge response source={source} not found")

        meta = dict(words=words)
        op = agent.monitor.submit(source, longrunning.OpTypes.challenge, metadata=meta)
        rep.status = falcon.HTTP_202
        rep.content_type = "application/json"
        rep.data = op.to_json().encode("utf-8")

        rep.status = falcon.HTTP_202

    @staticmethod
    def on_put(req, rep, source):
        """ Challenge PUT accept endpoint

        Parameters:
            req: falcon.Request HTTP request
            rep: falcon.Response HTTP response
            source: qb64 AID of of source of signed response to verify

        ---
        summary:  Mark challenge response exn message as signed
        description:  Mark challenge response exn message as signed
        tags:
           - Challenge/Response
        parameters:
          - in: path
            name: name
            schema:
              type: string
            required: true
            description: Human readable alias for the identifier to create
        requestBody:
            required: true
            content:
              application/json:
                schema:
                    description: Challenge response
                    properties:
                        aid:
                          type: string
                          description: aid of signer of accepted challenge response
                        said:
                          type: array
                          description:  SAID of challenge message signed
                          items:
                              type: string
        responses:
           202:
              description: Success submission of signed challenge/response
        """
        agent = req.context.agent
        body = req.get_media()
        if "said" not in body:
            raise falcon.HTTPBadRequest(description="challenge response acceptance requires 'aid' and 'said'")

        if source not in agent.hby.kevers:
            raise falcon.HTTPNotFound(description=f"challenge response source={source} not found")

        said = body["said"]
        saider = coring.Saider(qb64=said)
        agent.hby.db.chas.add(keys=(source,), val=saider)

        rep.status = falcon.HTTP_202


class ContactCollectionEnd:

    def on_get(self, req, rep):
        """ Contact plural GET endpoint

        Parameters:
            req: falcon.Request HTTP request
            rep: falcon.Response HTTP response
        ---
        summary:  Get list of contact information associated with remote identifiers
        description:  Get list of contact information associated with remote identifiers.  All
                      information is metadata and kept in local storage only
        tags:
           - Contacts
        parameters:
          - in: query
            name: group
            schema:
              type: string
            required: false
            description: field name to group results by
          - in: query
            name: filter_field
            schema:
               type: string
            description: field name to search
            required: false
          - in: query
            name: filter_value
            schema:
               type: string
            description: value to search for
            required: false
        responses:
           200:
              description: List of contact information for remote identifiers
        """
        # TODO:  Add support for sorting
        agent = req.context.agent
        group = req.params.get("group")
        field = req.params.get("filter_field")
        val = req.params.get("filter_value")

        if group is not None:
            data = dict()
            values = agent.org.values(group, val)
            for value in values:
                contacts = agent.org.find(group, value)
                self.authn(agent, contacts)
                data[value] = contacts

            rep.status = falcon.HTTP_200
            rep.data = json.dumps(data).encode("utf-8")

        elif field is not None:
            val = req.params.get("filter_value")
            if val is None:
                raise falcon.HTTPBadRequest(description="filter_value if required if field_field is specified")

            contacts = agent.org.find(field=field, val=val)
            self.authn(agent, contacts)
            rep.status = falcon.HTTP_200
            rep.data = json.dumps(contacts).encode("utf-8")

        else:
            data = []
            contacts = agent.org.list()

            for contact in contacts:
                aid = contact["id"]
                if aid in agent.hby.kevers and aid not in agent.hby.prefixes:
                    data.append(contact)

            self.authn(agent, data)
            rep.status = falcon.HTTP_200
            rep.data = json.dumps(data).encode("utf-8")

    @staticmethod
    def authn(agent, contacts):
        for contact in contacts:
            aid = contact['id']
            accepted = [saider.qb64 for saider in agent.hby.db.chas.get(keys=(aid,))]
            received = [saider.qb64 for saider in agent.hby.db.reps.get(keys=(aid,))]

            challenges = []
            for said in received:
                exn = agent.hby.db.exns.get(keys=(said,))
                challenges.append(dict(dt=exn.ked['dt'], words=exn.ked['a']['words'], said=said,
                                       authenticated=said in accepted))

            contact["challenges"] = challenges

            wellKnowns = []
            wkans = agent.hby.db.wkas.get(keys=(aid,))
            for wkan in wkans:
                wellKnowns.append(dict(url=wkan.url, dt=wkan.dt))

            contact["wellKnowns"] = wellKnowns


class ContactImageResourceEnd:

    @staticmethod
    def on_post(req, rep, prefix):
        """

        Parameters:
            req: falcon.Request HTTP request
            rep: falcon.Response HTTP response
            prefix: qb64 identifier prefix of contact to associate with image

        ---
         summary: Uploads an image to associate with identifier.
         description: Uploads an image to associate with identifier.
         tags:
            - Contacts
         parameters:
           - in: path
             name: prefix
             schema:
                type: string
             description: identifier prefix to associate image to
         requestBody:
             required: true
             content:
                image/jpg:
                  schema:
                    type: string
                    format: binary
                image/png:
                  schema:
                    type: string
                    format: binary
         responses:
           200:
              description: Image successfully uploaded

        """
        agent = req.context.agent
        if prefix not in agent.hby.kevers:
            raise falcon.HTTPNotFound(description=f"{prefix} is not a known identifier.")

        if req.content_length > 1000000:
            raise falcon.HTTPBadRequest(description="image too big to save")

        agent.org.setImg(pre=prefix, typ=req.content_type, stream=req.bounded_stream)
        rep.status = falcon.HTTP_202

    @staticmethod
    def on_get(req, rep, prefix):
        """ Contact image GET endpoint

        Parameters:
            req: falcon.Request HTTP request
            rep: falcon.Response HTTP response
            prefix: qb64 identifier prefix of contact information to get

       ---
        summary:  Get contact image for identifer prefix
        description:  Get contact image for identifer prefix
        tags:
           - Contacts
        parameters:
          - in: path
            name: prefix
            schema:
              type: string
            required: true
            description: qb64 identifier prefix of contact image to get
        responses:
           200:
              description: Contact information successfully retrieved for prefix
              content:
                  image/jpg:
                    schema:
                        description: Image
                        type: binary
           404:
              description: No contact information found for prefix
        """
        agent = req.context.agent
        if prefix not in agent.hby.kevers:
            raise falcon.HTTPNotFound(description=f"{prefix} is not a known identifier.")

        data = agent.org.getImgData(pre=prefix)
        if data is None:
            raise falcon.HTTPNotFound(description=f"no image available for {prefix}.")

        rep.status = falcon.HTTP_200
        rep.set_header('Content-Type', data["type"])
        rep.set_header('Content-Length', data["length"])
        rep.stream = agent.org.getImg(pre=prefix)


class ContactResourceEnd:

    @staticmethod
    def on_get(req, rep, prefix):
        """ Contact GET endpoint

        Parameters:
            req: falcon.Request HTTP request
            rep: falcon.Response HTTP response
            prefix: qb64 identifier prefix of contact information to get

       ---
        summary:  Get contact information associated with single remote identifier
        description:  Get contact information associated with single remote identifier.  All
                      information is meta-data and kept in local storage only
        tags:
           - Contacts
        parameters:
          - in: path
            name: prefix
            schema:
              type: string
            required: true
            description: qb64 identifier prefix of contact to get
        responses:
           200:
              description: Contact information successfully retrieved for prefix
           404:
              description: No contact information found for prefix
        """
        agent = req.context.agent
        if prefix not in agent.hby.kevers:
            raise falcon.HTTPNotFound(description=f"{prefix} is not a known identifier.")

        contact = agent.org.get(prefix)
        if contact is None:
            raise falcon.HTTPNotFound(description="NOT FOUND")

        rep.status = falcon.HTTP_200
        rep.data = json.dumps(contact).encode("utf-8")

    @staticmethod
    def on_post(req, rep, prefix):
        """ Contact plural GET endpoint

        Parameters:
            req: falcon.Request HTTP request
            rep: falcon.Response HTTP response
            prefix: human readable name of identifier to replace contact information

       ---
        summary:  Create new contact information for an identifier
        description:  Creates new information for an identifier, overwriting all existing
                      information for that identifier
        tags:
           - Contacts
        parameters:
          - in: path
            name: prefix
            schema:
              type: string
            required: true
            description: qb64 identifier prefix to add contact metadata to
        requestBody:
            required: true
            content:
              application/json:
                schema:
                    description: Contact information
                    type: object

        responses:
           200:
              description: Updated contact information for remote identifier
           400:
              description: Invalid identifier used to update contact information
           404:
              description: Prefix not found in identifier contact information
        """
        agent = req.context.agent
        body = req.get_media()
        if prefix not in agent.hby.kevers:
            raise falcon.HTTPNotFound(description="{prefix} is not a known identifier.  oobi required before contact "
                                                  "information")

        if prefix in agent.hby.prefixes:
            raise falcon.HTTPBadRequest(description=f"{prefix} is a local identifier, contact information only for "
                                                    f"remote identifiers")

        if "id" in body:
            del body["id"]

        if agent.org.get(prefix):
            raise falcon.HTTPBadRequest(description=f"contact data for {prefix} already exists")

        agent.org.replace(prefix, body)
        contact = agent.org.get(prefix)

        rep.status = falcon.HTTP_200
        rep.data = json.dumps(contact).encode("utf-8")

    @staticmethod
    def on_put(req, rep, prefix):
        """ Contact PUT endpoint

        Parameters:
            req: falcon.Request HTTP request
            rep: falcon.Response HTTP response
            prefix: qb64 identifier to update contact information

        ---
        summary:  Update provided fields in contact information associated with remote identifier prefix
        description:  Update provided fields in contact information associated with remote identifier prefix.  All
                      information is metadata and kept in local storage only
        tags:
           - Contacts
        parameters:
          - in: path
            name: prefix
            schema:
              type: string
            required: true
            description: qb64 identifier prefix to add contact metadata to
        requestBody:
            required: true
            content:
              application/json:
                schema:
                    description: Contact information
                    type: object

        responses:
           200:
              description: Updated contact information for remote identifier
           400:
              description: Invalid identifier used to update contact information
           404:
              description: Prefix not found in identifier contact information
        """
        agent = req.context.agent
        body = req.get_media()
        if prefix not in agent.hby.kevers:
            raise falcon.HTTPNotFound(
                description=f"{prefix} is not a known identifier.  oobi required before contact information")

        if prefix in agent.hby.prefixes:
            raise falcon.HTTPBadRequest(
                description=f"{prefix} is a local identifier, contact information only for remote identifiers")

        if "id" in body:
            del body["id"]

        agent.org.update(prefix, body)
        contact = agent.org.get(prefix)

        rep.status = falcon.HTTP_200
        rep.data = json.dumps(contact).encode("utf-8")

    @staticmethod
    def on_delete(req, rep, prefix):
        """ Contact plural GET endpoint

        Parameters:
            req: falcon.Request HTTP request
            rep: falcon.Response HTTP response
            prefix: qb64 identifier prefix to delete contact information

        ---
        summary:  Delete contact information associated with remote identifier
        description:  Delete contact information associated with remote identifier
        tags:
           - Contacts
        parameters:
          - in: path
            name: prefix
            schema:
              type: string
            required: true
            description: qb64 identifier prefix of contact to delete
        responses:
           202:
              description: Contact information successfully deleted for prefix
           404:
              description: No contact information found for prefix
        """
        agent = req.context.agent
        deleted = agent.org.rem(prefix)
        if not deleted:
            raise falcon.HTTPNotFound(description=f"no contact information to delete for {prefix}")

        rep.status = falcon.HTTP_202


class GroupMemberCollectionEnd:

    @staticmethod
    def on_get(req, rep, name):
        agent = req.context.agent

        hab = agent.hby.habByName(name)
        if hab is None:
            raise falcon.errors.HTTPNotFound(description=f"invalid alias {name}")

        if not isinstance(hab, habbing.SignifyGroupHab):
            raise falcon.HTTPBadRequest(description="members endpoint only available for group AIDs")

        smids = hab.db.signingMembers(hab.pre)
        rmids = hab.db.rotationMembers(hab.pre)

        signing = []
        for smid in smids:
            ends = hab.endsFor(smid)
            signing.append(dict(aid=smid, ends=ends))

        rotation = []
        for rmid in rmids:
            ends = hab.endsFor(rmid)
            rotation.append(dict(aid=rmid, ends=ends))

        data = dict(signing=signing, rotation=rotation)
        rep.status = falcon.HTTP_200
        rep.data = json.dumps(data).encode("utf-8")



# -*- encoding: utf-8 -*-
"""
KERIA
keria.app.credentialing module

services and endpoint for ACDC credential managements
"""
import json
from dataclasses import asdict

import falcon
from keri import kering
from keri.app import signing
from keri.app.habbing import SignifyGroupHab
from keri.core import coring, scheming, serdering
from keri.db import dbing
from keri.db.dbing import dgKey
from keri.vdr import viring

from keria.core import httping, longrunning


def loadEnds(app, identifierResource):
    schemaColEnd = SchemaCollectionEnd()
    app.add_route("/schema", schemaColEnd)
    schemaResEnd = SchemaResourceEnd()
    app.add_route("/schema/{said}", schemaResEnd)

    registryEnd = RegistryCollectionEnd(identifierResource)
    app.add_route("/identifiers/{name}/registries", registryEnd)

    registryResEnd = RegistryResourceEnd()
    app.add_route("/identifiers/{name}/registries/{registryName}", registryResEnd)

    credentialCollectionEnd = CredentialCollectionEnd(identifierResource)
    app.add_route("/identifiers/{name}/credentials", credentialCollectionEnd)
    
    credentialResourceEnd = CredentialResourceEnd()
    app.add_route("/credentials/{said}", credentialResourceEnd)
    credentialResourceDelEnd = CredentialResourceDeleteEnd(identifierResource)
    app.add_route("/identifiers/{name}/credentials/{said}", credentialResourceDelEnd)

    queryCollectionEnd = CredentialQueryCollectionEnd()
    app.add_route("/credentials/query", queryCollectionEnd)


class RegistryCollectionEnd:
    """
    ReST API for admin of credential issuance and revocation registries

    """

    def __init__(self, identifierResource):
        """

        Parameters:
            identifierResource (IdentifierResourceEnd): endpoint class for creating rotation and interaction events

        """
        self.identifierResource = identifierResource

    @staticmethod
    def on_get(req, rep, name):
        """  Registries GET endpoint

        Parameters:
            req: falcon.Request HTTP request
            rep: falcon.Response HTTP response
            name (str): human readable name for AID

        ---
        summary: List credential issuance and revocation registies
        description: List credential issuance and revocation registies
        tags:
           - Registries
        responses:
           200:
              description:  array of current credential issuance and revocation registies

        """
        agent = req.context.agent

        hab = agent.hby.habByName(name)
        if hab is None:
            raise falcon.HTTPNotFound(description="name is not a valid reference to an identifier")

        res = []
        for name, registry in agent.rgy.regs.items():
            if registry.regk not in registry.tevers:  # defensive programming for a registry not being fully committed
                continue

            if registry.hab.pre == hab.pre:
                rd = dict(
                    name=registry.name,
                    regk=registry.regk,
                    pre=registry.hab.pre,
                    state=asdict(registry.tever.state())
                )
                res.append(rd)

        rep.status = falcon.HTTP_200
        rep.content_type = "application/json"
        rep.data = json.dumps(res).encode("utf-8")

    def on_post(self, req, rep, name):
        """  Registries POST endpoint

        Parameters:
            req: falcon.Request HTTP request
            rep: falcon.Response HTTP response
            name (str): AID of Hab to load credentials for

        ---
        summary: Request to create a credential issuance and revocation registry
        description: Request to create a credential issuance and revocation registry
        tags:
           - Registries
        requestBody:
            required: true
            content:
              application/json:
                schema:
                  type: object
                  properties:
                    name:
                      type: string
                      description: name of the new registry
                    alias:
                      type: string
                      description: name of identifier to associate as the issuer of the new credential registry
                    toad:
                      type: integer
                      description: Backer receipt threshold
                    nonce:
                      type: string
                      description: qb64 encoded ed25519 random seed for registry
                    noBackers:
                      type: boolean
                      required: False
                      description: True means to not allow seperate backers from identifier's witnesses.
                    baks:
                      type: array
                      items:
                         type: string
                      description: List of qb64 AIDs of witnesses to be used for the new group identifier.
                    estOnly:
                      type: boolean
                      required: false
                      default: false
                      description: True means to not allow interaction events to anchor credential events.
        responses:
           202:
              description:  registry inception request has been submitted

        """
        agent = req.context.agent
        body = req.get_media()

        rname = httping.getRequiredParam(body, "name")
        ked = httping.getRequiredParam(body, "vcp")
        vcp = serdering.SerderKERI(sad=ked)

        ked = httping.getRequiredParam(body, "ixn")
        ixn = serdering.SerderKERI(sad=ked)

        hab = agent.hby.habByName(name)
        if hab is None:
            raise falcon.HTTPNotFound(description="alias is not a valid reference to an identifier")

        registry = agent.rgy.makeSignifyRegistry(name=rname, prefix=hab.pre, regser=vcp)

        if hab.kever.estOnly:
            op = self.identifierResource.rotate(agent, name, body)
        else:
            op = self.identifierResource.interact(agent, name, body)

        anchor = dict(i=registry.regk, s="0", d=registry.regk)
        # Create registry long running OP that embeds the above received OP or Serder.

        seqner = coring.Seqner(sn=ixn.sn)
        prefixer = coring.Prefixer(qb64=ixn.pre)
        agent.registrar.incept(hab, registry, prefixer=prefixer, seqner=seqner, saider=coring.Saider(qb64=ixn.said))
        op = agent.monitor.submit(hab.kever.prefixer.qb64, longrunning.OpTypes.registry,
                                  metadata=dict(anchor=anchor, depends=op))

        rep.status = falcon.HTTP_202
        rep.data = op.to_json().encode("utf-8")


class RegistryResourceEnd:

    @staticmethod
    def on_get(req, rep, name, registryName):
        """  Registry Resource GET endpoint

        Parameters:
            req: falcon.Request HTTP request
            rep: falcon.Response HTTP response
            name (str): human readable name for AID
            registryName(str): human readable name for registry

        ---
        summary: Get a single credential issuance and revocation registy
        description: Get a single credential issuance and revocation registy
        tags:
           - Registries
        responses:
           200:
              description:  credential issuance and revocation registy

        """
        agent = req.context.agent

        hab = agent.hby.habByName(name)
        if hab is None:
            raise falcon.HTTPNotFound(description=f"{name} is not a valid reference to an identifier")

        registry = agent.rgy.registryByName(registryName)
        if registry is None:
            raise falcon.HTTPNotFound(description=f"{registryName} is not a valid reference to a credential registry")

        if not registry.hab.pre == hab.pre:
            raise falcon.HTTPNotFound(description=f"{registryName} is not a valid registry for AID {name}")

        rd = dict(
            name=registry.name,
            regk=registry.regk,
            pre=registry.hab.pre,
            state=asdict(registry.tever.state())
        )
        rep.status = falcon.HTTP_200
        rep.content_type = "application/json"
        rep.data = json.dumps(rd).encode("utf-8")

    @staticmethod
    def on_put(req, rep, name, registryName):
        """  Registry Resource PUT endpoint

        Parameters:
            req: falcon.Request HTTP request
            rep: falcon.Response HTTP response
            name (str): human readable name for AID
            registryName(str): human readable name for registry or its SAID

        ---
        summary: Get a single credential issuance and revocation registy
        description: Get a single credential issuance and revocation registy
        tags:
           - Registries
        responses:
           200:
              description:  credential issuance and revocation registy

        """
        agent = req.context.agent

        hab = agent.hby.habByName(name)
        if hab is None:
            raise falcon.HTTPNotFound(description=f"{name} is not a valid reference to an identifier")

        body = req.get_media()
        if 'name' not in body:
            raise falcon.HTTPBadRequest(description="'name' is required in body")

        name = body['name']
        if agent.rgy.registryByName(name) is not None:
            raise falcon.HTTPBadRequest(description=f"{name} is already in use for a registry")

        registry = agent.rgy.registryByName(registryName)
        if registry is None:
            if registryName in agent.rgy.regs:  # Check to see if the registryName parameter is a SAID
                registry = agent.rgy.regs[registryName]
            else:
                regk = registryName
                key = dgKey(regk, regk)
                raw = agent.rgy.reger.getTvt(key=key)
                if raw is None:
                    raise falcon.HTTPNotFound(
                        description=f"{registryName} is not a valid reference to a credential registry")

                regser = serdering.SerderKERI(raw=bytes(raw))
                registry = agent.rgy.makeSignifyRegistry(name, hab.pre, regser)

        regord = viring.RegistryRecord(registryKey=registry.regk, prefix=hab.pre)
        agent.rgy.reger.regs.pin(keys=(name,), val=regord)
        registry.name = name

        rd = dict(
            name=registry.name,
            regk=registry.regk,
            pre=registry.hab.pre,
            state=asdict(registry.tever.state())
        )
        rep.status = falcon.HTTP_200
        rep.content_type = "application/json"
        rep.data = json.dumps(rd).encode("utf-8")


class SchemaResourceEnd:

    @staticmethod
    def on_get(req, rep, said):
        """ Schema GET endpoint

        Parameters:
            req: falcon.Request HTTP request
            rep: falcon.Response HTTP response
            said: qb64 self-addressing identifier of schema to load

       ---
        summary:  Get schema JSON of specified schema
        description:  Get schema JSON of specified schema
        tags:
           - Schema
        parameters:
          - in: path
            name: said
            schema:
              type: string
            required: true
            description: qb64 self-addressing identifier of schema to get
        responses:
           200:
              description: Schema JSON successfully returned
           404:
              description: No schema found for SAID
        """
        agent = req.context.agent
        schemer = agent.hby.db.schema.get(keys=(said,))
        if schemer is None:
            raise falcon.HTTPNotFound(description="Schema not found")

        data = schemer.sed
        rep.status = falcon.HTTP_200
        rep.data = json.dumps(data).encode("utf-8")


class SchemaCollectionEnd:

    @staticmethod
    def on_get(req, rep):
        """ Schema GET plural endpoint

        Parameters:
            req: falcon.Request HTTP request
            rep: falcon.Response HTTP response

       ---
        summary:  Get schema JSON of all schema
        description:  Get schema JSON of all schema
        tags:
           - Schema
        responses:
           200:
              description: Array of all schema JSON
        """
        agent = req.context.agent

        data = []
        for said, schemer in agent.hby.db.schema.getItemIter():
            data.append(schemer.sed)

        rep.status = falcon.HTTP_200
        rep.data = json.dumps(data).encode("utf-8")


class CredentialQueryCollectionEnd:
    """ This class provides a collection endpoint for creating credential queries.

    I fully admit that the semantics here are a big stretch.  I would rather have this as a GET against the
    credential collection endpoint, but the nature of the complicated input to this endpoint dictate a BODY
    and certain client libraries (and possibly reverse proxies) don't support a BODY in a GET request.  So
    I'm moving the credential query code to this endpoint class and mapping to `.../credentials/queries` and
    making it a post against that path and calling it "creating a creaential query".  Meh.

    """

    @staticmethod
    def on_post(req, rep):
        """ Credentials GET endpoint

        Parameters:
            req: falcon.Request HTTP request
            rep: falcon.Response HTTP response

        ---
        summary:  List credentials in credential store (wallet)
        description: List issued or received credentials current verified
        tags:
           - Credentials
        parameters:
           - in: path
             name: aid
             schema:
               type: string
             required: true
             description: identifier to load credentials for
           - in: query
             name: type
             schema:
                type: string
             description:  type of credential to return, [issued|received]
             required: true
           - in: query
             name: schema
             schema:
                type: string
             description:  schema to filter by if provided
             required: false
        responses:
           200:
              description: Credential list.
              content:
                  application/json:
                    schema:
                        description: Credentials
                        type: array
                        items:
                           type: object

        """
        agent = req.context.agent
        try:
            body = req.get_media()
            if "filter" in body:
                filtr = body["filter"]
            else:
                filtr = {}

            if "sort" in body:
                sort = body["sort"]
            else:
                sort = None

            if "skip" in body:
                skip = body["skip"]
            else:
                skip = 0

            if "limit" in body:
                limit = body["limit"]
            else:
                limit = 25
        except falcon.HTTPError:
            filtr = {}
            sort = {}
            skip = 0
            limit = 25

        cur = agent.seeker.find(filtr=filtr, sort=sort, skip=skip, limit=limit)
        saids = [coring.Saider(qb64=said) for said in cur]
        creds = agent.rgy.reger.cloneCreds(saids=saids, db=agent.hby.db)

        end = skip + (len(creds) - 1) if len(creds) > 0 else 0
        rep.set_header("Accept-Ranges", "credentials")
        rep.set_header("Content-Range", f"credentials {skip}-{end}/{limit}")

        rep.status = falcon.HTTP_200
        rep.content_type = "application/json"
        rep.data = json.dumps(creds).encode("utf-8")


class CredentialCollectionEnd:

    def __init__(self, identifierResource):
        """

        Parameters:
            identifierResource (IdentifierResourceEnd): endpoint class for creating rotation and interaction events

        """
        self.identifierResource = identifierResource

    def on_post(self, req, rep, name):
        """ Initiate a credential issuance

        Parameters:
            req: falcon.Request HTTP request
            rep: falcon.Response HTTP response
            name (str): human readable alias for AID to use as issuer

        ---
        summary: Perform credential issuance
        description: Perform credential issuance
        tags:
           - Credentials
        parameters:
          - in: path
            name: alias
            schema:
              type: string
            required: true
            description: Human readable alias for the identifier to create
        requestBody:
            required: true
            content:
              application/json:
                schema:
                  type: object
                  properties:
                    registry:
                      type: string
                      description: Alias of credential issuance/revocation registry (aka status)
                    recipient:
                      type: string
                      description: AID of credential issuance/revocation recipient
                    schema:
                      type: string
                      description: SAID of credential schema being issued
                    rules:
                      type: object
                      description: Rules section (Ricardian contract) for credential being issued
                    source:
                      type: object
                      description: ACDC edge or edge group for chained credentials
                      properties:
                         d:
                            type: string
                            description: SAID of reference chain
                         s:
                            type: string
                            description: SAID of reference chain schema
                    credentialData:
                      type: object
                      description: dynamic map of values specific to the schema
                    private:
                      type: boolean
                      description: flag to inidicate this credential should support privacy preserving presentations
        responses:
           200:
              description: Credential issued.
              content:
                  application/json:
                    schema:
                        description: Credential
                        type: object

        """
        agent = req.context.agent

        body = req.get_media()
        hab = agent.hby.habByName(name)
        if hab is None:
            raise falcon.HTTPNotFound(description="name is not a valid reference to an identifier")

        creder = serdering.SerderACDC(sad=httping.getRequiredParam(body, "acdc"))
        iserder = serdering.SerderKERI(sad=httping.getRequiredParam(body, "iss"))
        if "ixn" in body:
            anc = serdering.SerderKERI(sad=httping.getRequiredParam(body, "ixn"))
        else:
            anc = serdering.SerderKERI(sad=httping.getRequiredParam(body, "rot"))

        regk = iserder.ked['ri']
        if regk not in agent.rgy.tevers:
            raise falcon.HTTPNotFound(description=f"issue against invalid registry SAID {regk}")

        if hab.kever.estOnly:
            op = self.identifierResource.rotate(agent, name, body)
        else:
            op = self.identifierResource.interact(agent, name, body)

        try:
            agent.credentialer.validate(creder)
            agent.registrar.issue(regk, iserder, anc)
            agent.credentialer.issue(creder=creder, serder=iserder)
            op = agent.monitor.submit(hab.kever.prefixer.qb64, longrunning.OpTypes.credential,
                                      metadata=dict(ced=creder.sad, depends=op))

        except kering.ConfigurationError as e:
            rep.status = falcon.HTTP_400
            rep.text = e.args[0]
            return

        rep.status = falcon.HTTP_200
        rep.data = op.to_json().encode("utf-8")


class CredentialResourceEnd:
    def __init__(self):
        """

        """

    @staticmethod
    def on_get(req, rep, said):
        """ Credentials GET endpoint

        Parameters:
            req: falcon.Request HTTP request
            rep: falcon.Response HTTP response
            said (str): SAID of credential to export

        ---
        summary:  Export credential and all supporting cryptographic material
        description: Export credential and all supporting cryptographic material
        tags:
           - Credentials
        parameters:
           - in: path
             name: aid
             schema:
               type: string
             required: true
             description: The identifier to create
           - in: path
             name: said
             schema:
               type: string
             required: true
             description: SAID of credential to get
        responses:
           200:
              description: Credential export.
              content:
                  application/json+cesr:
                    schema:
                        description: Credential
                        type: object

        """
        agent = req.context.agent
        accept = req.get_header("accept")
        if accept == "application/json+cesr":
            rep.content_type = "application/json+cesr"
            data = CredentialResourceEnd.outputCred(agent.hby, agent.rgy, said)
        else:
            rep.content_type = "application/json"
            creds = agent.rgy.reger.cloneCreds([coring.Saider(qb64=said)], db=agent.hby.db)
            if not creds:
                raise falcon.HTTPNotFound(description=f"credential for said {said} not found.")

            data = json.dumps(creds[0]).encode("utf-8")

        rep.status = falcon.HTTP_200
        rep.data = bytes(data)

    @staticmethod
    def outputCred(hby, rgy, said):
        out = bytearray()
        creder, prefixer, seqner, saider = rgy.reger.cloneCred(said=said)
        chains = creder.edge or dict()
        saids = []
        for key, source in chains.items():
            if key == 'd':
                continue

            if not isinstance(source, dict):
                continue

            saids.append(source['n'])

        for said in saids:
            out.extend(CredentialResourceEnd.outputCred(hby, rgy, said))

        issr = creder.issuer
        for msg in hby.db.clonePreIter(pre=issr):
            serder = serdering.SerderKERI(raw=msg)
            atc = msg[serder.size:]
            out.extend(serder.raw)
            out.extend(atc)

        if "i" in creder.attrib:
            subj = creder.attrib["i"]
            for msg in hby.db.clonePreIter(pre=subj):
                serder = serdering.SerderKERI(raw=msg)
                atc = msg[serder.size:]
                out.extend(serder.raw)
                out.extend(atc)

        if creder.regi is not None:
            for msg in rgy.reger.clonePreIter(pre=creder.regi):
                serder = serdering.SerderKERI(raw=msg)
                atc = msg[serder.size:]
                out.extend(serder.raw)
                out.extend(atc)

            for msg in rgy.reger.clonePreIter(pre=creder.said):
                serder = serdering.SerderKERI(raw=msg)
                atc = msg[serder.size:]
                out.extend(serder.raw)
                out.extend(atc)

        out.extend(signing.serialize(creder, prefixer, seqner, saider))

        return out


class CredentialResourceDeleteEnd:
    def __init__(self, identifierResource):
            """

            Parameters:
                identifierResource (IdentifierResourceEnd): endpoint class for creating rotation and interaction events

            """
            self.identifierResource = identifierResource

    def on_delete(self, req, rep, name, said):
        """ Initiate a credential revocation

        Parameters:
            req: falcon.Request HTTP request
            rep: falcon.Response HTTP response
            name (str): human readable alias for AID to use as issuer
            said (str): SAID of credential to revoke

        RequestBody:
            rev (str): serialized revocation event
            ixn (str): serialized interaction event
            rot (str): serialized rotation event
            sigs (list): list of signatures for the revocation event
        ---
        summary: Perform credential revocation
        description: Perform credential revocation
        """

        agent = req.context.agent

        body = req.get_media()
        hab = agent.hby.habByName(name)
        if hab is None:
            raise falcon.HTTPNotFound(description="name is not a valid reference to an identifier")

        rserder = serdering.SerderKERI(sad=httping.getRequiredParam(body, "rev"))

        regk = rserder.ked['ri']
        if regk not in agent.rgy.tevers:
            raise falcon.HTTPNotFound(description=f"revocation against invalid registry SAID {regk}")

        try:
            agent.rgy.reger.cloneCreds([coring.Saider(qb64=said)], db=agent.hby.db)
        except:
            raise falcon.HTTPNotFound(description=f"credential for said {said} not found.")

        if hab.kever.estOnly:
            op = self.identifierResource.rotate(agent, name, body)
            anc = httping.getRequiredParam(body, "rot")
        else:
            op = self.identifierResource.interact(agent, name, body)
            anc = httping.getRequiredParam(body, "ixn")

        try:
            agent.registrar.revoke(regk, rserder, anc)
        except Exception as e:
            raise falcon.HTTPBadRequest(description=f"invalid revocation event.")

        rep.status = falcon.HTTP_200
        rep.data = op.to_json().encode("utf-8")


def signPaths(hab, pather, sigers):
    """ Sign the SAD or SAIDs with the keys from the Habitat.

    Sign the SADs or SAIDs of the SADs as identified by the paths.

    Parameters:
        hab (Habitat): environment used to sign the SAD
        pather (Pather): Pather for the signatures
        sigers (list): list of signatures over the paths

    Returns:
        list: pathed signature tuples

    """

    sadsigers = []

    prefixer, seqner, saider, indices = signing.transSeal(hab)
    sadsigers.append((pather, prefixer, seqner, saider, sigers))

    return sadsigers


class Registrar:

    def __init__(self, agentHab, hby, rgy, counselor, witDoer, witPub, verifier):
        self.hby = hby
        self.agentHab = agentHab
        self.rgy = rgy
        self.counselor = counselor
        self.witDoer = witDoer
        self.witPub = witPub
        self.verifier = verifier

    def incept(self, hab, registry, prefixer=None, seqner=None, saider=None):
        """

        Parameters:
            hab (Hab): human readable name for the registry
            registry (SignifyRegistry): qb64 identifier prefix of issuing identifier in control of this registry
            prefixer (Prefixer):
            seqner (Seqner): sequence number class of anchoring event
            saider (Saider): SAID class of anchoring event

        Returns:
            Registry:  created registry

        """
        rseq = coring.Seqner(sn=0)
        if not isinstance(hab, SignifyGroupHab):

            seqner = coring.Seqner(sn=hab.kever.sner.num)
            saider = coring.Saider(qb64=hab.kever.serder.said)
            registry.anchorMsg(pre=registry.regk, regd=registry.regd, seqner=seqner, saider=saider)
            self.witDoer.msgs.append(dict(pre=hab.pre, sn=seqner.sn))
            self.rgy.reger.tpwe.add(keys=(registry.regk, rseq.qb64), val=(hab.kever.prefixer, seqner, saider))

        else:
            print("Waiting for TEL registry vcp event mulisig anchoring event")
            self.rgy.reger.tmse.add(keys=(registry.regk, rseq.qb64, registry.regd), val=(prefixer, seqner, saider))

    def issue(self, regk, iserder, anc):
        """
        Create and process the credential issuance TEL events on the given registry

        Parameters:
            regk (str): qb64 identifier prefix of the credential registry
            iserder (Serder): TEL issuance event
            anc (Serder): Anchoring KEL event

        """
        registry = self.rgy.regs[regk]
        registry.processEvent(serder=iserder)
        hab = registry.hab

        vcid = iserder.ked["i"]
        rseq = coring.Seqner(snh=iserder.ked["s"])

        if not isinstance(hab, SignifyGroupHab):  # not a multisig group
            seqner = coring.Seqner(sn=hab.kever.sner.num)
            saider = coring.Saider(qb64=hab.kever.serder.said)
            registry.anchorMsg(pre=vcid, regd=iserder.said, seqner=seqner, saider=saider)

            print("Waiting for TEL event witness receipts")
            self.witDoer.msgs.append(dict(pre=hab.pre, sn=seqner.sn))
            self.rgy.reger.tpwe.add(keys=(vcid, rseq.qb64), val=(hab.kever.prefixer, seqner, saider))
            return vcid, rseq.sn

        else:  # multisig group hab
            sn = anc.sn
            said = anc.said

            prefixer = coring.Prefixer(qb64=hab.pre)
            seqner = coring.Seqner(sn=sn)
            saider = coring.Saider(qb64=said)

            print(f"Waiting for TEL iss event mulisig anchoring event {seqner.sn}")
            self.rgy.reger.tmse.add(keys=(vcid, rseq.qb64, iserder.said), val=(prefixer, seqner, saider))
            return vcid, rseq.sn

    def revoke(self, regk, rserder, anc):
        """
        Create and process the credential revocation TEL events on the given registry

        Parameters:
            regk (str): qb64 identifier prefix of the credential registry
            rserder (Serder): TEL revocation event
            anc (Serder): KEL anchoring event
        """
        registry = self.rgy.regs[regk]
        registry.processEvent(serder=rserder)
        hab = registry.hab

        vcid = rserder.ked["i"]
        rseq = coring.Seqner(snh=rserder.ked["s"])

        if not isinstance(hab, SignifyGroupHab):

            seqner = coring.Seqner(sn=hab.kever.sner.num)
            saider = coring.Saider(qb64=hab.kever.serder.said)
            registry.anchorMsg(pre=vcid, regd=rserder.said, seqner=seqner, saider=saider)

            print("Waiting for TEL event witness receipts")
            self.witDoer.msgs.append(dict(pre=hab.pre, sn=seqner.sn))

            self.rgy.reger.tpwe.add(keys=(vcid, rseq.qb64), val=(hab.kever.prefixer, seqner, saider))
            return vcid, rseq.sn
        else:
            serder = serdering.SerderKERI(sad=anc)
            sn = serder.sn
            said = serder.said

            prefixer = coring.Prefixer(qb64=hab.pre)
            seqner = coring.Seqner(sn=sn)
            saider = coring.Saider(qb64=said)

            self.counselor.start(prefixer=prefixer, seqner=seqner, saider=saider, ghab=hab)

            print(f"Waiting for TEL rev event mulisig anchoring event {seqner.sn}")
            self.rgy.reger.tmse.add(keys=(vcid, rseq.qb64, rserder.said), val=(prefixer, seqner, saider))
            return vcid, rseq.sn

    def complete(self, pre, sn=0):
        """ Determine if registry event (inception, issuance, revocation, etc.) is finished validation

        Parameters:
            pre (str): qb64 identifier of registry event
            sn (int): integer sequence number of regsitry event

        Returns:
            bool: True means event has completed and is commited to database
        """

        seqner = coring.Seqner(sn=sn)
        said = self.rgy.reger.ctel.get(keys=(pre, seqner.qb64))
        return said is not None

    def processEscrows(self):
        """
        Process credential registry anchors:

        """
        self.processWitnessEscrow()
        self.processMultisigEscrow()
        self.processDiseminationEscrow()

    def processWitnessEscrow(self):
        """
        Process escrow of group multisig events that do not have a full compliment of receipts
        from witnesses yet.  When receipting is complete, remove from escrow and cue up a message
        that the event is complete.

        """
        for (regk, snq), (prefixer, seqner, saider) in self.rgy.reger.tpwe.getItemIter():  # partial witness escrow
            kever = self.hby.kevers[prefixer.qb64]
            dgkey = dbing.dgKey(prefixer.qb64b, saider.qb64)

            # Load all the witness receipts we have so far
            wigs = self.hby.db.getWigs(dgkey)
            if kever.wits:
                if len(wigs) == len(kever.wits):  # We have all of them, this event is finished
                    hab = self.hby.habs[prefixer.qb64]
                    witnessed = False
                    for cue in self.witDoer.cues:
                        if cue["pre"] == hab.pre and cue["sn"] == seqner.sn:
                            witnessed = True

                    if not witnessed:
                        continue
                else:
                    continue

            rseq = coring.Seqner(qb64=snq)
            self.rgy.reger.tpwe.rem(keys=(regk, snq))

            self.rgy.reger.tede.add(keys=(regk, rseq.qb64), val=(prefixer, seqner, saider))

    def processMultisigEscrow(self):
        """
        Process escrow of group multisig events that do not have a full compliment of receipts
        from witnesses yet.  When receipting is complete, remove from escrow and cue up a message
        that the event is complete.

        """
        for (regk, snq, regd), (prefixer, seqner, saider) in self.rgy.reger.tmse.getItemIter():  # multisig escrow
            try:
                if not self.counselor.complete(prefixer, seqner, saider):
                    continue
            except kering.ValidationError:
                self.rgy.reger.tmse.rem(keys=(regk, snq, regd))
                continue

            rseq = coring.Seqner(qb64=snq)

            # Anchor the message, registry or otherwise
            key = dbing.dgKey(regk, regd)
            sealet = seqner.qb64b + saider.qb64b
            self.rgy.reger.putAnc(key, sealet)

            self.rgy.reger.tmse.rem(keys=(regk, snq, regd))
            self.rgy.reger.tede.add(keys=(regk, rseq.qb64), val=(prefixer, seqner, saider))

    def processDiseminationEscrow(self):
        for (regk, snq), (prefixer, seqner, saider) in self.rgy.reger.tede.getItemIter():  # group multisig escrow
            rseq = coring.Seqner(qb64=snq)
            dig = self.rgy.reger.getTel(key=dbing.snKey(pre=regk, sn=rseq.sn))
            if dig is None:
                continue

            self.rgy.reger.tede.rem(keys=(regk, snq))

            tevt = bytearray()
            for msg in self.rgy.reger.clonePreIter(pre=regk, fn=rseq.sn):
                tevt.extend(msg)

            print(f"Sending TEL events to witnesses")
            # Fire and forget the TEL event to the witnesses.  Consumers will have to query
            # to determine when the Witnesses have received the TEL events.
            self.witPub.msgs.append(dict(pre=prefixer.qb64, msg=tevt))
            self.rgy.reger.ctel.put(keys=(regk, rseq.qb64), val=saider)  # idempotent


class Credentialer:

    def __init__(self, agentHab, hby, rgy, registrar, verifier, notifier):
        self.agentHab = agentHab
        self.hby = hby
        self.rgy = rgy
        self.registrar = registrar
        self.verifier = verifier
        self.notifier = notifier

    def validate(self, creder):
        """

        Args:
            creder (Creder): creder object representing the credential to validate

        Returns:
            bool: true if credential is valid against a known schema

        """
        schema = creder.sad['s']
        scraw = self.verifier.resolver.resolve(schema)
        if not scraw:
            raise kering.ConfigurationError("Credential schema {} not found.  It must be loaded with data oobi before "
                                            "issuing credentials".format(schema))

        schemer = scheming.Schemer(raw=scraw)
        try:
            schemer.verify(creder.raw)
        except kering.ValidationError as ex:
            raise kering.ConfigurationError(f"Credential schema validation failed for {schema}: {ex}")

        return True

    def issue(self, creder, serder):
        """ Issue the credential creder and handle witness propagation and communication

        Parameters:
            creder (Creder): Credential object to issue
            serder (Serder): KEL or TEL anchoring event

        """
        prefixer = coring.Prefixer(qb64=serder.pre)
        seqner = coring.Seqner(sn=serder.sn)

        self.rgy.reger.cmse.put(keys=(creder.said, seqner.qb64), val=creder)

        try:
            self.verifier.processCredential(creder=creder, prefixer=prefixer, seqner=seqner,
                                            saider=coring.Saider(qb64=serder.said))
        except kering.MissingRegistryError:
            pass

    def processCredentialMissingSigEscrow(self):
        for (said, snq), creder in self.rgy.reger.cmse.getItemIter():
            rseq = coring.Seqner(qb64=snq)
            if not self.registrar.complete(pre=said, sn=rseq.sn):
                continue

            saider = self.rgy.reger.saved.get(keys=said)
            if saider is None:
                continue

            # Remove from this escrow
            self.rgy.reger.cmse.rem(keys=(said, snq))

            # place in escrow to diseminate to other if witnesser and if there is an issuee
            self.rgy.reger.ccrd.put(keys=(creder.said,), val=creder)

    def complete(self, said):
        return self.rgy.reger.ccrd.get(keys=(said,)) is not None

    def processEscrows(self):
        """
        Process credential registry anchors:

        """
        self.processCredentialMissingSigEscrow()



# -*- encoding: utf-8 -*-
"""
KERIA
keria.core.longrunning module

"""
import datetime
from collections import namedtuple
from dataclasses import dataclass, asdict

import falcon
import json
from dataclasses_json import dataclass_json
from keri import kering
from keri.app.oobiing import Result
from keri.core import eventing, coring, serdering
from keri.db import dbing, koming
from keri.help import helping

# long running operationt types
Typeage = namedtuple("Tierage", 'oobi witness delegation group query registry credential endrole challenge exchange '
                                'submit done')

OpTypes = Typeage(oobi="oobi", witness='witness', delegation='delegation', group='group', query='query',
                  registry='registry', credential='credential', endrole='endrole', challenge='challenge',
                  exchange='exchange', submit='submit', done='done')


@dataclass_json
@dataclass
class Status:
    code: int
    message: str
    details: dict = None


@dataclass_json
@dataclass
class Operation:
    name: str
    metadata: dict
    done: bool = False
    error: Status = None
    response: dict = None


@dataclass
class Op:
    oid: str
    type: str
    start: str
    metadata: dict


class Operator(dbing.LMDBer):
    TailDirPath = "keri/opr"
    AltTailDirPath = ".keri/opr"
    TempPrefix = "keri_ops_"

    def __init__(self, name="opr", headDirPath=None, reopen=True, **kwa):
        """

        Parameters:
            headDirPath:
            perm:
            reopen:
            kwa:
        """
        self.ops = None
        self.msgs = None

        super(Operator, self).__init__(name=name, headDirPath=headDirPath, reopen=reopen, **kwa)

    def reopen(self, **kwa):
        """  Reopen database and initialize sub-dbs
        """
        super(Operator, self).reopen(**kwa)

        # Long running operations, keyed by "name" which is f"{type}.{oid}"
        self.ops = koming.Komer(db=self, subkey='opr.', schema=Op, )

        return self.env


class Monitor:
    """  Monitoring and garbage collecting long running operations

    Attributes:
        hby (Habery): identifier database environment
        opr(Operator): long running operations database
        swain(Sealer): Delegation processes tracker

    """

    def __init__(self, hby, swain, counselor=None, registrar=None, exchanger=None, credentialer=None, opr=None,
                 submitter=None, temp=False):
        """ Create long running operation monitor

        Parameters:
            hby (Habery): identifier database environment
            swain(Sealer): Delegation processes tracker
            counselor (Counselor): Group processes tracker
            registrar (Registrar): Registry processes tracker
            exchanger (Exchanger): Exchange processes tracker
            credentialer (Credentialer): Credential processes tracker
            opr (Operator): long running operations database
            submitter (Submitter): Submit processes tracker

        """
        self.hby = hby
        self.swain = swain
        self.counselor = counselor
        self.registrar = registrar
        self.exchanger = exchanger
        self.credentialer = credentialer
        self.opr = opr if opr is not None else Operator(name=hby.name, temp=temp)
        self.submitter = submitter

    def submit(self, oid, typ, metadata=None):
        """  Submit a new long running operation to track

        parameters:
            oid (str): opaque identifier of the target of the long running operation (type specific)
            typ (Typeage): long running operation type
            metadata (dict): additional metadata for the long running operation (type specific)

        Returns:
            Operation: the initial long running operation status object

        """
        if typ not in OpTypes:
            raise kering.InvalidValueError(f"{typ} not a valid long running operation type")

        name = f"{typ}.{oid}"
        op = Op(oid=oid, type=typ, start=helping.nowIso8601(), metadata=metadata)

        # Overwrite any existing long running operation of this type for this resource.
        # resets the clock basically
        self.opr.ops.pin(keys=(name,), val=op)

        # Return Operation with full status check in case its already finished.
        return self.get(name)

    def get(self, name):
        if (op := self.opr.ops.get(keys=(name,))) is None:
            return None

        operation = self.status(op)

        return operation

    def getOperations(self, type=None):
        """ Return list of long running opterations, optionally filtered by type """
        ops = self.opr.ops.getItemIter()
        if type != None:
            ops = filter(lambda i: i[1].type == type, ops)

        def get_status(op):
            try:
                return self.status(op)
            except Exception as err:
                # self.status may throw an exception.
                # Handling error by returning an operation with error status
                return Operation(
                    name=f"{op.type}.{op.oid}",
                    metadata=op.metadata,
                    done=True,
                    error=Status(code=500, message=f"{err}"))

        return [get_status(op) for (_, op) in ops]

    def rem(self, name):
        """ Remove tracking of the long running operation represented by name """
        return self.opr.ops.rem(keys=(name,))

    def status(self, op):
        """  Calculate the status of an operation.

        Base on the type of an operation, determine the current status of the operation, including loading
        any availabie error messages if the operation failed of the sucessful result of the operation which
        will be the same that would be returned by the original endpoint if the operation were not long running

        Parameters:
            op (Op): database storage for long running operation

        Returns:
            Operation: The status of the operation, including any errors or results as appropriate.

        """

        operation = Operation(
            name=f"{op.type}.{op.oid}",
            metadata=op.metadata,
        )

        if op.type in (OpTypes.witness,):
            if op.oid not in self.hby.kevers:
                raise kering.ValidationError(f"long running {op.type} operation identifier {op.oid} not found")

            if "sn" not in op.metadata:
                raise kering.ValidationError(f"invalid long running {op.type} operation, metadata missing 'sn' field")

            sn = op.metadata["sn"]
            kever = self.hby.kevers[op.oid]
            sdig = self.hby.db.getKeLast(key=dbing.snKey(pre=kever.prefixer.qb64b, sn=sn))
            if sdig is not None:

                dgkey = dbing.dgKey(kever.prefixer.qb64b, bytes(sdig))
                wigs = self.hby.db.getWigs(dgkey)

                if len(wigs) >= kever.toader.num:
                    evt = self.hby.db.getEvt(dbing.dgKey(pre=kever.prefixer.qb64, dig=bytes(sdig)))
                    serder = serdering.SerderKERI(raw=bytes(evt))
                    operation.done = True
                    operation.response = serder.ked

                else:
                    start = helping.fromIso8601(op.start)
                    dtnow = helping.nowUTC()
                    if (dtnow - start) > datetime.timedelta(seconds=eventing.Kevery.TimeoutPWE):
                        operation.done = True
                        operation.error = Status(code=408,  # Using HTTP error codes here for lack of a better alternative
                                                 message=f"long running {op.type} for {op.oid} operation timed out before "
                                                         f"receiving sufficient witness receipts")
                    else:
                        operation.done = False

            else:
                operation.done = False

        elif op.type in (OpTypes.oobi,):
            if "oobi" not in op.metadata:
                raise kering.ValidationError("invalid OOBI long running operation, missing oobi")

            oobi = op.metadata["oobi"]
            obr = self.hby.db.roobi.get(keys=(oobi,))
            if obr is None:
                operation.done = False
            elif obr.state == Result.resolved:
                operation.done = True
                if obr.cid and obr.cid in self.hby.kevers:
                    kever = self.hby.kevers[obr.cid]
                    operation.response = asdict(kever.state())
                else:
                    operation.response = dict(oobi=oobi)

            elif obr.state == Result.failed:
                operation.done = True
                operation.failed = Status(code=500,
                                          message=f"resolving OOBI {op.oid} failed")
            else:
                operation.done = False

        elif op.type in (OpTypes.delegation, ):
            if op.oid not in self.hby.kevers:
                raise kering.ValidationError(f"long running {op.type} operation identifier {op.oid} not found")

            if "sn" not in op.metadata:
                raise kering.ValidationError(f"invalid long running {op.type} operation, metadata missing 'sn' field")

            kever = self.hby.kevers[op.oid]
            sn = op.metadata["sn"]
            seqner = coring.Seqner(sn=sn)
            sdig = self.hby.db.getKeLast(key=dbing.snKey(pre=op.oid, sn=sn))

            if self.swain.complete(kever.prefixer, seqner):
                evt = self.hby.db.getEvt(dbing.dgKey(pre=kever.prefixer.qb64, dig=bytes(sdig)))
                serder = serdering.SerderKERI(raw=bytes(evt))

                operation.done = True
                operation.response = serder.ked
            else:
                operation.done = False

        elif op.type in (OpTypes.group, ):
            if "sn" not in op.metadata:
                raise kering.ValidationError(f"invalid long running {op.type} operation, metadata missing 'sn' field")

            prefixer = coring.Prefixer(qb64=op.oid)
            seqner = coring.Seqner(sn=op.metadata["sn"])

            if self.counselor.complete(prefixer, seqner):
                sdig = self.hby.db.getKeLast(key=dbing.snKey(pre=op.oid, sn=seqner.sn))
                evt = self.hby.db.getEvt(dbing.dgKey(pre=prefixer.qb64, dig=bytes(sdig)))
                serder = serdering.SerderKERI(raw=bytes(evt))

                operation.done = True
                operation.response = serder.ked
            else:
                operation.done = False

        elif op.type in (OpTypes.query, ):
            if op.oid not in self.hby.kevers:
                operation.done = False

            else:
                kever = self.hby.kevers[op.oid]
                if "sn" in op.metadata:
                    sn = int(op.metadata['sn'], 16)
                    if kever.sn >= sn:
                        operation.done = True
                        operation.response = asdict(kever.state())
                    else:
                        operation.done = False
                elif "anchor" in op.metadata:
                    anchor = op.metadata["anchor"]
                    if self.hby.db.findAnchoringSealEvent(op.oid, seal=anchor) is not None:
                        operation.done = True
                        operation.response = asdict(kever.state())
                    else:
                        operation.done = False
                else:
                    ksn = None
                    for (_, saider) in self.hby.db.knas.getItemIter(keys=(op.oid,)):
                        ksn = self.hby.db.ksns.get(keys=(saider.qb64,))
                        break

                    if ksn and ksn.d == kever.serder.said:
                        operation.done = True
                        operation.response = asdict(kever.state())
                    else:
                        operation.done = False

        elif op.type in (OpTypes.registry, ):
            if op.oid not in self.hby.kevers:
                raise kering.ValidationError(
                    f"long running {op.type} operation identifier {op.oid} not found")
            if "anchor" not in op.metadata:
                raise kering.ValidationError(
                    f"invalid long running {op.type} operation, metadata missing 'anchor' field")

            anchor = op.metadata["anchor"]
            if self.hby.db.findAnchoringSealEvent(op.oid, seal=anchor) is not None:
                operation.done = True
                operation.response = dict(anchor=anchor)
            else:
                operation.done = False

        elif op.type in (OpTypes.credential,):
            if "ced" not in op.metadata:
                raise kering.ValidationError(
                    f"invalid long running {op.type} operation, metadata missing 'ced' field")

            ced = op.metadata["ced"]
            if self.credentialer.complete(ced['d']):
                operation.done = True
                operation.response = dict(ced=ced)
            else:
                operation.done = False

        elif op.type in (OpTypes.exchange,):
            if "said" not in op.metadata:
                raise kering.ValidationError(
                    f"invalid long running {op.type} operation, metadata missing 'said' field")

            said = op.metadata["said"]
            if self.exchanger.complete(said):
                operation.done = True
                operation.response = dict(said=said)
            else:
                operation.done = False

        elif op.type in (OpTypes.endrole, ):
            if "cid" not in op.metadata or "role" not in op.metadata or "eid" not in op.metadata:
                raise kering.ValidationError(
                    f"invalid long running {op.type} operation, metadata missing 'cid'/'role'/'eid' field")

            cid = op.metadata['cid']
            role = op.metadata['role']
            eid = op.metadata['eid']

            end = self.hby.db.ends.get(keys=(cid, role, eid))
            if end and (end.enabled or end.allowed):
                saider = self.hby.db.eans.get(keys=(cid, role, eid))
                serder = self.hby.db.rpys.get(keys=(saider.qb64,))
                operation.done = True
                operation.response = serder.ked
            else:
                operation.done = False

        elif op.type in (OpTypes.challenge,):
            if op.oid not in self.hby.kevers:
                operation.done = False

            if "words" not in op.metadata:
                raise kering.ValidationError(
                    f"invalid long running {op.type} operation, metadata missing 'words' field")

            found = False
            words = op.metadata["words"]
            saiders = self.hby.db.reps.get(keys=(op.oid,))
            for saider in saiders:
                exn = self.hby.db.exns.get(keys=(saider.qb64,))
                if words == exn.ked['a']['words']:
                    found = True
                    break

            if found:
                operation.done = True
                operation.response = dict(exn=exn.ked)
            else:
                operation.done = False
        elif op.type in (OpTypes.submit, ):
            if "pre" not in op.metadata:
                raise kering.ValidationError(
                    f"invalid long running {op.type} operation, metadata missing 'pre' field")

            pre = op.metadata['pre']            
            self.submitter.msgs.append(dict(pre=pre))

            operation.done = True

        elif op.type in (OpTypes.done, ):
            operation.done = True
            operation.response = op.metadata["response"]

        else:
            operation.done = True
            operation.error = Status(code=404,  # Using HTTP error codes here for lack of a better alternative
                                     message=f"long running operation type {op.type} unknown")

        return operation


class OperationCollectionEnd:
    @staticmethod
    def on_get(req, rep):
        """ Get list of long running operations 
        
        Parameters:
            req (Request):  Falcon HTTP Request object
            rep (Response): Falcon HTTP Response object

        ---
        summary: Get list of long running operations
        parameters:
          - in: query
            name: type
            schema:
              type: string
            required: false
            description: filter list of long running operations by type
        responses:
            200:
              content:
                  application/json:
                    schema:
                        type: array

        """
        agent = req.context.agent
        type = req.params.get("type")
        ops = agent.monitor.getOperations(type=type)
        rep.data = json.dumps(ops, default=lambda o: o.to_dict()).encode("utf-8")
        rep.content_type = "application/json"
        rep.status = falcon.HTTP_200


class OperationResourceEnd:
    """ Single Resource REST endpoint for long running operations

    """

    @staticmethod
    def on_get(req, rep, name):
        """  GET single resource REST endpoint

        Parameters:
            req (Request):  Falcon HTTP Request object
            rep (Response): Falcon HTTP Response object
            name (str): Long running operation resource name to load

        """
        agent = req.context.agent
        if (operation := agent.monitor.get(name)) is None:
            raise falcon.HTTPNotFound(title=f"long running operation '{name}' not found")

        rep.content_type = "application/json"
        rep.data = operation.to_json().encode("utf-8")
        rep.status = falcon.HTTP_200

    @staticmethod
    def on_delete(req, rep, name):
        """ DELETE single resource REST endpoint

        Args:
            req (Request):  Falcon HTTP Request object
            rep (Response): Falcon HTTP Response object
            name (str): Long running operation resource name to load

        """

        agent = req.context.agent
        if agent.monitor.get(name) is None:
            raise falcon.HTTPNotFound(title=f"long running operation '{name}' not found")

        deleted = agent.monitor.rem(name)
        if deleted:
            rep.status = falcon.HTTP_204
        else:
            raise falcon.HTTPInternalServerError(f"unable to delete long running operation {name}")



# -*- encoding: utf-8 -*-
"""
KERIA
keria.core.httping module

"""

import falcon
from falcon.http_status import HTTPStatus

class HandleCORS(object):
    def process_request(self, req, resp):
        resp.set_header('Access-Control-Allow-Origin', '*')
        resp.set_header('Access-Control-Allow-Methods', '*')
        resp.set_header('Access-Control-Allow-Headers', '*')
        resp.set_header('Access-Control-Max-Age', 1728000)  # 20 days
        if req.method == 'OPTIONS':
            raise HTTPStatus(falcon.HTTP_200, body='\n')


def getRequiredParam(body, name):
    param = body.get(name)
    if param is None:
        raise falcon.HTTPBadRequest(description=f"required field '{name}' missing from request")

    return param


def parseRangeHeader(header, name, start=0, end=9):
    """ Parse the start and end requested range values, defaults are 0, 9

    Parameters:
        header(str):  HTTP Range header value
        name (str): range name to look for
        start (int): default start index
        end(int): default end index

    Returns:
        (start, end): tuple of start index and end index

    """

    if not header.startswith(f"{name}="):
        return start, end

    header = header.strip(f"{name}=")
    try:
        if header.startswith("-"):
            return start, int(header[1:])

        if header.endswith("-"):
            return int(header[:-1]), end

        vals = header.split("-")
        if not len(vals) == 2:
            return start, end

        return int(vals[0]), int(vals[1])
    except ValueError:
        return start, end





# -*- encoding: utf-8 -*-
"""
KERIA
keria.app.notifying module

"""
import json

import falcon
from keri.peer import exchanging

from keria.core import httping


def loadEnds(app):
    noteCol = NotificationCollectionEnd()
    app.add_route("/notifications", noteCol)
    noteRes = NotificationResourceEnd()
    app.add_route("/notifications/{said}", noteRes)


class NotificationCollectionEnd:

    @staticmethod
    def on_get(req, rep):
        """ Notification GET endpoint

        Parameters:
            req: falcon.Request HTTP request
            rep: falcon.Response HTTP response
        ---
        summary:  Get list of notifications for the controller of the agent
        description:  Get list of notifications for the controller of the agent.  Notifications will
                       be sorted by creation date/time
        parameters:
          - in: header
            name: Range
            schema:
              type: string
            required: false
            description: HTTP Range header syntax
            description: size of the result list.  Defaults to 25
        tags:
           - Notifications

        responses:
           200:
              description: List of contact information for remote identifiers
        """
        agent = req.context.agent

        rng = req.get_header("Range")
        if rng is None:
            rep.status = falcon.HTTP_200
            start = 0
            end = 24
        else:
            rep.status = falcon.HTTP_206
            start, end = httping.parseRangeHeader(rng, "notes")

        count = agent.notifier.getNoteCnt()
        notes = agent.notifier.getNotes(start=start, end=end)
        out = []
        for note in notes:
            attrs = note.pad
            out.append(attrs)

        end = start + (len(out) - 1) if len(out) > 0 else 0
        rep.set_header("Accept-Ranges", "notes")
        rep.set_header("Content-Range", f"notes {start}-{end}/{count}")
        rep.status = falcon.HTTP_200
        rep.data = json.dumps(out).encode("utf-8")


class NotificationResourceEnd:

    @staticmethod
    def on_put(req, rep, said):
        """ Notification PUT endpoint

        Parameters:
            req: falcon.Request HTTP request
            rep: falcon.Response HTTP response
            said: qb64 SAID of notification to mark as read

        ---
        summary:  Mark notification as read
        description:  Mark notification as read
        tags:
           - Notifications
        parameters:
          - in: path
            name: said
            schema:
              type: string
            required: true
            description: qb64 said of note to mark as read
        responses:
           202:
              description: Notification successfully marked as read for prefix
           404:
              description: No notification information found for SAID
        """
        agent = req.context.agent
        mared = agent.notifier.mar(said)
        if not mared:
            rep.status = falcon.HTTP_404
            rep.data = json.dumps(dict(msg=f"no notification to mark as read for {said}")).encode("utf-8")
            return

        rep.status = falcon.HTTP_202

    @staticmethod
    def on_delete(req, rep, said):
        """ Notification DELETE endpoint

        Parameters:
            req: falcon.Request HTTP request
            rep: falcon.Response HTTP response
            said: qb64 SAID of notification to delete

        ---
        summary:  Delete notification
        description:  Delete notification
        tags:
           - Notifications
        parameters:
          - in: path
            name: said
            schema:
              type: string
            required: true
            description: qb64 said of note to delete
        responses:
           202:
              description: Notification successfully deleted for prefix
           404:
              description: No notification information found for prefix
        """
        agent = req.context.agent
        deleted = agent.notifier.noter.rem(said)
        if not deleted:
            rep.status = falcon.HTTP_404
            rep.text = f"no notification to delete for {said}"
            return

        rep.status = falcon.HTTP_202



# -*- encoding: utf-8 -*-
"""
KERIA
keria.cli.keria.commands module

Witness command line interface
"""
import argparse
import logging

from keri import __version__
from keri import help
from keri.app import directing

from keria.app import agenting

d = "Runs KERI Signify Agent\n"
d += "\tExample:\nkli ahab\n"
parser = argparse.ArgumentParser(description=d)
parser.set_defaults(handler=lambda args: launch(args))
parser.add_argument('-V', '--version',
                    action='version',
                    version=__version__,
                    help="Prints out version of script runner.")
parser.add_argument('-a', '--admin-http-port',
                    dest="admin",
                    action='store',
                    default=3901,
                    help="Admin port number the HTTP server listens on. Default is 3901.")
parser.add_argument('-H', '--http',
                    action='store',
                    default=3902,
                    help="Local port number the HTTP server listens on. Default is 3902.")
parser.add_argument('-B', '--boot',
                    action='store',
                    default=3903,
                    help="Boot port number the Boot HTTP server listens on.  This port needs to be secured."
                         " Default is 3903.")
parser.add_argument('-n', '--name',
                    action='store',
                    default="keria",
                    help="Name of controller. Default is agent.")
parser.add_argument('--base', '-b', help='additional optional prefix to file location of KERI keystore',
                    required=False, default="")
parser.add_argument('--passcode', '-p', help='22 character encryption passcode for keystore (is not saved)',
                    dest="bran", default=None)  # passcode => bran
parser.add_argument('--config-file',
                    dest="configFile",
                    action='store',
                    default="",
                    help="configuration filename")
parser.add_argument("--config-dir",
                    dest="configDir",
                    action="store",
                    default=None,
                    help="directory override for configuration data")
parser.add_argument("--keypath", action="store", required=False, default=None,
                    help="TLS server private key file")
parser.add_argument("--certpath", action="store", required=False, default=None,
                    help="TLS server signed certificate (public key) file")
parser.add_argument("--cafilepath", action="store", required=False, default=None,
                    help="TLS server CA certificate chain")


def launch(args):
    help.ogler.level = logging.CRITICAL
    help.ogler.reopen(name=args.name, temp=True, clear=True)

    logger = help.ogler.getLogger()

    logger.info("******* Starting Agent for %s listening: admin/%s, http/%s "
                ".******", args.name, args.admin, args.http)

    runAgent(name=args.name,
             base=args.base,
             bran=args.bran,
             admin=int(args.admin),
             http=int(args.http),
             boot=int(args.boot),
             configFile=args.configFile,
             configDir=args.configDir,
             keypath=args.keypath,
             certpath=args.certpath,
             cafilepath=args.cafilepath)

    logger.info("******* Ended Agent for %s listening: admin/%s, http/%s"
                ".******", args.name, args.admin, args.http)


def runAgent(name="ahab", base="", bran="", admin=3901, http=3902, boot=3903, configFile=None,
             configDir=None, keypath=None, certpath=None, cafilepath=None, expire=0.0):
    """
    Setup and run a KERIA Agency
    """

    doers = []
    doers.extend(agenting.setup(name=name, base=base, bran=bran,
                                adminPort=admin,
                                httpPort=http,
                                bootPort=boot,
                                configFile=configFile,
                                configDir=configDir,
                                keypath=keypath,
                                certpath=certpath,
                                cafilepath=cafilepath))

    directing.runController(doers=doers, expire=expire)



# -*- encoding: utf-8 -*-
"""
KERIA
keria.app.grouping module

"""
import json

import falcon
from keri.app import habbing
from keri.core import coring, eventing, serdering
from keri.kering import SerializeError

from keria.core import httping, longrunning


def loadEnds(app):
    msrCol = MultisigRequestCollectionEnd()
    app.add_route("/identifiers/{name}/multisig/request", msrCol)
    joinCol = MultisigJoinCollectionEnd()
    app.add_route("/identifiers/{name}/multisig/join", joinCol)
    msrRes = MultisigRequestResourceEnd()
    app.add_route("/multisig/request/{said}", msrRes)


class MultisigRequestCollectionEnd:
    """ Collection endpoint class for creating mulisig exn requests from """

    @staticmethod
    def on_post(req, rep, name):
        """ POST method for multisig request collection

        Parameters:
            req (falcon.Request): HTTP request object
            rep (falcon.Response): HTTP response object
            name (str): AID of Hab to load credentials for

        """
        agent = req.context.agent

        body = req.get_media()

        # Get the hab
        hab = agent.hby.habByName(name)
        if hab is None:
            raise falcon.HTTPNotFound(description=f"alias={name} is not a valid reference to an identifier")

        # ...and make sure we're a Group
        if not isinstance(hab, habbing.SignifyGroupHab):
            raise falcon.HTTPBadRequest(description=f"hab for alias {name} is not a multisig")

        # grab all of the required parameters
        ked = httping.getRequiredParam(body, "exn")
        serder = serdering.SerderKERI(sad=ked)
        sigs = httping.getRequiredParam(body, "sigs")
        atc = httping.getRequiredParam(body, "atc")

        # create sigers from the edge signatures so we can messagize the whole thing
        sigers = [coring.Siger(qb64=sig) for sig in sigs]

        # create seal for the proper location to find the signatures
        kever = hab.mhab.kever
        seal = eventing.SealEvent(i=hab.mhab.pre, s="{:x}".format(kever.lastEst.s), d=kever.lastEst.d)

        ims = eventing.messagize(serder=serder, sigers=sigers, seal=seal)
        ims.extend(atc.encode("utf-8"))  # add the pathed attachments
        # make a copy and parse
        agent.hby.psr.parseOne(ims=bytearray(ims))
        # now get rid of the event so we can pass it as atc to send
        del ims[:serder.size]

        smids = hab.db.signingMembers(pre=hab.pre)
        smids.remove(hab.mhab.pre)

        agent.exchanges.append(dict(said=serder.said, pre=hab.pre, rec=smids, topic='multisig'))

        rep.status = falcon.HTTP_200
        rep.data = json.dumps(serder.ked).encode("utf-8")


class MultisigJoinCollectionEnd:
    """ Collection endpoint class for creating mulisig exn requests from """

    @staticmethod
    def on_post(req, rep, name):
        """ POST method for multisig request collection

        Parameters:
            req (falcon.Request): HTTP request object
            rep (falcon.Response): HTTP response object
            name (str): AID of Hab to load credentials for

        """
        agent = req.context.agent

        # Get the hab
        hab = agent.hby.habByName(name)
        if hab is not None:
            raise falcon.HTTPBadRequest(description=f"attempt to create identifier with an already used alias={name}")

        agent = req.context.agent
        body = req.get_media()

        # Get the rot, sigs and recipients  from the request
        rot = httping.getRequiredParam(body, "rot")
        sigs = httping.getRequiredParam(body, "sigs")

        # Get group specific values
        gid = httping.getRequiredParam(body, "gid")
        smids = httping.getRequiredParam(body, "smids")
        rmids = httping.getRequiredParam(body, "rmids")

        both = list(set(smids + (rmids or [])))
        for recp in both:  # Have to verify we already know all the recipients.
            if recp not in agent.hby.kevers:
                agent.hby.deleteHab(name=name)
                raise falcon.HTTPBadRequest(description=f"attempt to merge with unknown AID={recp}")

        sigers = [coring.Siger(qb64=sig) for sig in sigs]
        verfers = [coring.Verfer(qb64=k) for k in rot['k']]
        digers = [coring.Diger(qb64=n) for n in rot['n']]

        mhab = None
        for mid in both:
            if mid in agent.hby.habs:
                mhab = agent.hby.habs[mid]
                break

        if mhab is None:
            raise falcon.HTTPBadRequest(description="Invalid multisig group rotation request,"
                                                    " signing member list must contain a local identifier'")

        hab = agent.hby.joinSignifyGroupHab(gid, name=name, mhab=mhab, smids=smids, rmids=rmids)
        try:
            hab.make(serder=serdering.SerderKERI(sad=rot), sigers=sigers)
            agent.inceptGroup(pre=gid, mpre=mhab.pre, verfers=verfers, digers=digers)
        except (ValueError, SerializeError) as e:
            agent.hby.deleteHab(name=name)
            raise falcon.HTTPBadRequest(description=f"{e.args[0]}")

        serder = serdering.SerderKERI(sad=rot)
        agent.groups.append(dict(pre=hab.pre, serder=serder, sigers=sigers, smids=smids, rmids=rmids))
        op = agent.monitor.submit(serder.pre, longrunning.OpTypes.group, metadata=dict(sn=0))

        rep.content_type = "application/json"
        rep.status = falcon.HTTP_202
        rep.data = op.to_json().encode("utf-8")


class MultisigRequestResourceEnd:
    """ Resource endpoint class for getting full data for a mulisig exn request from a notification """

    @staticmethod
    def on_get(req, rep, said):
        """ GET method for multisig resources

        Parameters:
            req (falcon.Request): HTTP request object
            rep (falcon.Response): HTTP response object
            said (str): qb64 SAID of EXN multisig message.

        """
        agent = req.context.agent
        exn = agent.hby.db.exns.get(keys=(said,))
        if exn is None:
            raise falcon.HTTPNotFound(description=f"no multisig request with said={said} found")

        route = exn.ked['r']
        if not route.startswith("/multisig"):
            raise falcon.HTTPBadRequest(f"invalid mutlsig conversation with said={said}")

        payload = exn.ked['a']
        match route.split("/"):
            case ["", "multisig", "icp"]:
                pass
            case ["", "multisig", "rot"]:
                pass
            case ["", "multisig", *_]:
                gid = payload["gid"]
                if gid not in agent.hby.habs:
                    raise falcon.HTTPBadRequest(f"multisig request for non-local group pre={gid}")

        esaid = exn.ked['e']['d']
        exns = agent.mux.get(esaid=esaid)

        for d in exns:
            exn = d['exn']
            serder = serdering.SerderKERI(sad=exn)

            route = serder.ked['r']
            payload = serder.ked['a']
            match route.split("/"):
                case ["", "multisig", "icp"]:
                    pass
                case ["", "multisig", "rot"]:
                    gid = payload["gid"]
                    if gid in agent.hby.habs:
                        ghab = agent.hby.habs[gid]
                        d['groupName'] = ghab.name
                        d['memberName'] = ghab.mhab.name

                case ["", "multisig", "vcp"]:
                    gid = payload["gid"]
                    ghab = agent.hby.habs[gid]
                    d['groupName'] = ghab.name
                    d['memberName'] = ghab.mhab.name

                    sender = serder.ked['i']
                    if (c := agent.org.get(sender)) is not None:
                        d['sender'] = c['alias']
                case ["", "multisig", "iss"]:
                    gid = payload["gid"]
                    ghab = agent.hby.habs[gid]
                    d['groupName'] = ghab.name
                    d['memberName'] = ghab.mhab.name

                    sender = serder.ked['i']
                    if (c := agent.org.get(sender)) is not None:
                        d['sender'] = c['alias']

        rep.status = falcon.HTTP_200
        rep.data = json.dumps(exns).encode("utf-8")



# -*- encoding: utf-8 -*-
"""
KERIA
keria.app.ipexing module

services and endpoint for IPEX message managements
"""
import json

import falcon
from keri.app import habbing
from keri.core import coring, eventing, serdering
from keri.peer import exchanging

from keria.core import httping, longrunning


def loadEnds(app):
    admitColEnd = IpexAdmitCollectionEnd()
    app.add_route("/identifiers/{name}/ipex/admit", admitColEnd)
    grantColEnd = IpexGrantCollectionEnd()
    app.add_route("/identifiers/{name}/ipex/grant", grantColEnd)


class IpexAdmitCollectionEnd:

    @staticmethod
    def on_post(req, rep, name):
        """  Registries GET endpoint

        Parameters:
            req: falcon.Request HTTP request
            rep: falcon.Response HTTP response
            name (str): human readable name for AID

        ---
        summary: List credential issuance and revocation registies
        description: List credential issuance and revocation registies
        tags:
           - Registries
        responses:
           200:
              description:  array of current credential issuance and revocation registies

        """
        agent = req.context.agent
        # Get the hab
        hab = agent.hby.habByName(name)
        if hab is None:
            raise falcon.HTTPNotFound(description=f"alias={name} is not a valid reference to an identifier")

        body = req.get_media()

        ked = httping.getRequiredParam(body, "exn")
        sigs = httping.getRequiredParam(body, "sigs")
        atc = httping.getRequiredParam(body, "atc")
        rec = httping.getRequiredParam(body, "rec")

        route = ked['r']

        match route:
            case "/ipex/admit":
                op = IpexAdmitCollectionEnd.sendAdmit(agent, hab, ked, sigs, rec)
            case "/multisig/exn":
                op = IpexAdmitCollectionEnd.sendMultisigExn(agent, hab, ked, sigs, atc, rec)
            case _:
                raise falcon.HTTPBadRequest(description=f"invalid message route {route}")

        rep.status = falcon.HTTP_200
        rep.data = op.to_json().encode("utf-8")

    @staticmethod
    def sendAdmit(agent, hab, ked, sigs, rec):
        for recp in rec:  # Have to verify we already know all the recipients.
            if recp not in agent.hby.kevers:
                raise falcon.HTTPBadRequest(description=f"attempt to send to unknown AID={recp}")

        # use that data to create th Serder and Sigers for the exn
        serder = serdering.SerderKERI(sad=ked)
        sigers = [coring.Siger(qb64=sig) for sig in sigs]

        # Now create the stream to send, need the signer seal
        kever = hab.kever
        seal = eventing.SealEvent(i=hab.pre, s="{:x}".format(kever.lastEst.s), d=kever.lastEst.d)

        ims = eventing.messagize(serder=serder, sigers=sigers, seal=seal)

        # make a copy and parse
        agent.hby.psr.parseOne(ims=bytearray(ims))

        # now get rid of the event so we can pass it as atc to send
        del ims[:serder.size]

        agent.exchanges.append(dict(said=serder.said, pre=hab.pre, rec=rec, topic='credential'))
        agent.admits.append(dict(said=ked['d'], pre=hab.pre))

        return agent.monitor.submit(serder.pre, longrunning.OpTypes.exchange, metadata=dict(said=serder.said))

    @staticmethod
    def sendMultisigExn(agent, hab, ked, sigs, atc, rec):
        if not isinstance(hab, habbing.SignifyGroupHab):
            raise falcon.HTTPBadRequest(description=f"attempt to send multisig message with non-group AID={hab.pre}")

        for recp in rec:  # Have to verify we already know all the recipients.
            if recp not in agent.hby.kevers:
                raise falcon.HTTPBadRequest(description=f"attempt to send to unknown AID={recp}")

        embeds = ked['e']
        admitked = embeds['exn']
        if admitked['r'] != "/ipex/admit":
            raise falcon.HTTPBadRequest(description=f"invalid route for embedded ipex admit {admitked['r']}")

        # Have to add the atc to the end... this will be Pathed signatures for embeds
        if not atc:
            raise falcon.HTTPBadRequest(description=f"attachment missing for ACDC, unable to process request.")

        # use that data to create th Serder and Sigers for the exn
        serder = serdering.SerderKERI(sad=ked)
        sigers = [coring.Siger(qb64=sig) for sig in sigs]

        # Now create the stream to send, need the signer seal
        kever = hab.mhab.kever
        seal = eventing.SealEvent(i=hab.mhab.pre, s="{:x}".format(kever.lastEst.s), d=kever.lastEst.d)

        ims = eventing.messagize(serder=serder, sigers=sigers, seal=seal)
        ims.extend(atc.encode("utf-8"))  # add the pathed attachments

        # make a copy and parse
        agent.hby.psr.parseOne(ims=bytearray(ims))
        agent.exchanges.append(dict(said=serder.said, pre=hab.pre, rec=rec, topic='credential'))

        exn, pathed = exchanging.cloneMessage(agent.hby, serder.said)
        if not exn:
            raise falcon.HTTPBadRequest(description=f"invalid exn request message {serder.said}")

        grant, _ = exchanging.cloneMessage(agent.hby, admitked['p'])
        if grant is None:
            raise falcon.HTTPBadRequest(description=f"attempt to admit an invalid grant {admitked['p']}")

        embeds = grant.ked['e']
        acdc = embeds["acdc"]
        issr = acdc['i']

        serder = serdering.SerderKERI(sad=admitked)
        ims = bytearray(serder.raw) + pathed['exn']
        agent.hby.psr.parseOne(ims=ims)
        agent.exchanges.append(dict(said=serder.said, pre=hab.pre, rec=[issr], topic="credential"))
        agent.admits.append(dict(said=admitked['d'], pre=hab.pre))

        return agent.monitor.submit(serder.pre, longrunning.OpTypes.exchange, metadata=dict(said=serder.said))


class IpexGrantCollectionEnd:

    @staticmethod
    def on_post(req, rep, name):
        """  Registries GET endpoint

        Parameters:
            req: falcon.Request HTTP request
            rep: falcon.Response HTTP response
            name (str): human readable name for AID

        ---
        summary: List credential issuance and revocation registies
        description: List credential issuance and revocation registies
        tags:
           - Registries
        responses:
           200:
              description:  array of current credential issuance and revocation registies

        """
        agent = req.context.agent
        hab = agent.hby.habByName(name)
        if hab is None:
            raise falcon.HTTPNotFound(description=f"alias={name} is not a valid reference to an identifier")

        body = req.get_media()

        ked = httping.getRequiredParam(body, "exn")
        sigs = httping.getRequiredParam(body, "sigs")
        atc = httping.getRequiredParam(body, "atc")
        rec = httping.getRequiredParam(body, "rec")

        route = ked['r']

        match route:
            case "/ipex/grant":
                op = IpexGrantCollectionEnd.sendGrant(agent, hab, ked, sigs, atc, rec)
            case "/multisig/exn":
                op = IpexGrantCollectionEnd.sendMultisigExn(agent, hab, ked, sigs, atc, rec)
            case _:
                raise falcon.HTTPBadRequest(description=f"invalid route {route}")

        rep.status = falcon.HTTP_200
        rep.data = op.to_json().encode("utf-8")

    @staticmethod
    def sendGrant(agent, hab, ked, sigs, atc, rec):
        for recp in rec:  # Have to verify we already know all the recipients.
            if recp not in agent.hby.kevers:
                raise falcon.HTTPBadRequest(description=f"attempt to send to unknown AID={recp}")

        # use that data to create th Serder and Sigers for the exn
        serder = serdering.SerderKERI(sad=ked)
        sigers = [coring.Siger(qb64=sig) for sig in sigs]

        # Now create the stream to send, need the signer seal
        kever = hab.kever
        seal = eventing.SealEvent(i=hab.pre, s="{:x}".format(kever.lastEst.s), d=kever.lastEst.d)

        ims = eventing.messagize(serder=serder, sigers=sigers, seal=seal)
        ims = ims + atc.encode("utf-8")

        # make a copy and parse
        agent.hby.psr.parseOne(ims=bytearray(ims))

        # now get rid of the event so we can pass it as atc to send
        del ims[:serder.size]

        agent.exchanges.append(dict(said=serder.said, pre=hab.pre, rec=rec, topic='credential'))
        agent.grants.append(dict(said=ked['d'], pre=hab.pre, rec=rec))

        return agent.monitor.submit(serder.pre, longrunning.OpTypes.exchange, metadata=dict(said=serder.said))

    @staticmethod
    def sendMultisigExn(agent, hab, ked, sigs, atc, rec):
        if not isinstance(hab, habbing.SignifyGroupHab):
            raise falcon.HTTPBadRequest(description=f"attempt to send multisig message with non-group AID={hab.pre}")

        for recp in rec:  # Have to verify we already know all the recipients.
            if recp not in agent.hby.kevers:
                raise falcon.HTTPBadRequest(description=f"attempt to send to unknown AID={recp}")

        embeds = ked['e']
        grant = embeds['exn']
        if grant['r'] != "/ipex/grant":
            raise falcon.HTTPBadRequest(description=f"invalid route for embedded ipex grant {ked['r']}")

        # use that data to create th Serder and Sigers for the exn
        serder = serdering.SerderKERI(sad=ked)
        sigers = [coring.Siger(qb64=sig) for sig in sigs]

        # Now create the stream to send, need the signer seal
        kever = hab.mhab.kever
        seal = eventing.SealEvent(i=hab.mhab.pre, s="{:x}".format(kever.lastEst.s), d=kever.lastEst.d)

        ims = eventing.messagize(serder=serder, sigers=sigers, seal=seal)

        ims.extend(atc.encode("utf-8"))  # add the pathed attachments

        # make a copy and parse
        agent.hby.psr.parseOne(ims=bytearray(ims))
        agent.exchanges.append(dict(said=serder.said, pre=hab.pre, rec=rec, topic='credential'))
        holder = grant['a']['i']

        exn, pathed = exchanging.cloneMessage(agent.hby, serder.said)
        if not exn:
            raise falcon.HTTPBadRequest(description=f"invalid exn request message {serder.said}")

        serder = serdering.SerderKERI(sad=grant)
        ims = bytearray(serder.raw) + pathed['exn']
        agent.hby.psr.parseOne(ims=ims)
        agent.exchanges.append(dict(said=serder.said, pre=hab.pre, rec=[holder], topic="credential"))
        agent.grants.append(dict(said=grant['d'], pre=hab.pre, rec=[holder]))

        return agent.monitor.submit(serder.pre, longrunning.OpTypes.exchange, metadata=dict(said=serder.said))
